package com.jalios.jcms;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.text.DateFormat;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Logger;
import org.jdom.Element;

import com.jalios.jcms.analytics.AnalyticsManager;
import com.jalios.jcms.analytics.EventData;
import com.jalios.jcms.analytics.agent.AccountReportAgent;
import com.jalios.jcms.caddy.CaddyComparator;
import com.jalios.jcms.comparator.BasicComparator;
import com.jalios.jcms.comparator.ComparatorManager;
import com.jalios.jcms.context.JcmsJspContext;
import com.jalios.jcms.db.DBData;
import com.jalios.jcms.db.DBDataRevision;
import com.jalios.jcms.db.DBDataRevisionManager;
import com.jalios.jcms.db.HibernateUtil;
import com.jalios.jcms.db.HistorizedDBData;
import com.jalios.jcms.dbmember.DBMember;
import com.jalios.jcms.extradbdata.ExtraDBDataManager;
import com.jalios.jcms.i18n.I18nUtil;
import com.jalios.jcms.mashup.ExportUtil;
import com.jalios.jcms.mashup.ImportConstants;
import com.jalios.jcms.mashup.ImportOptions;
import com.jalios.jcms.mashup.ImportUtil;
import com.jalios.jcms.policy.PortalPolicyFilter;
import com.jalios.jcms.uicomponent.TTCardManager;
import com.jalios.jcms.widget.WidgetUtil;
import com.jalios.jstore.BasicStorable;
import com.jalios.jstore.Storable;
import com.jalios.util.DateUtil;
import com.jalios.util.LangProperties;
import com.jalios.util.LangPropertyArgument;
import com.jalios.util.ReverseComparator;
import com.jalios.util.Util;
import com.jalios.util.XmlUtil;

/**
 * This abstract class is the super class of all the persistent
 * classes of a JCMS Channel.
 * 
 * @author Olivier Dedieu <olivier.dedieu@jalios.com>
 * @author Olivier Jaquemet <olivier.jaquemet@jalios.com>
 * @version $Revision: 98408 $
 */
public abstract class Data extends BasicStorable implements JcmsConstants, ImportConstants, LangPropertyArgument {
  public static final String REVISION = "$Revision: 98408 $";

  protected static final Channel channel = Channel.getChannel();
  private static final Logger logger = Logger.getLogger(Data.class);

  /**
   * Extra info to key to override default display URL String.
   * <p>
   * Use your custom display URL String as the value (need to be done only once).
   * <p>
   * Example : <br>
   * 
   * <pre>data.setExtraInfo(Data.DISPLAY_URL_EXTRA_INFO, "customDisplay.jsp?id=" + data.getId());</pre>
   * Later, any invokation of <code>data.getDisplayUrl()</code> returns <code>"customDisplay.jsp?id=..."</code>.
   */
  public static final String DISPLAY_URL_EXTRA_INFO = "com.jalios.jcms.data.DISPLAY_URL_EXTRA_INFO";
  
  private static String IMPORT_ID = "id";
  private static String IMPORT_DATE = "date";
  private static String IMPORT_MDATE = "mdate";
  private static String IMPORT_BATCH = "batch";
  private static String IMPORT_UPDATED = "updated";
  private static String IMPORT_SIGNATURE = "signature";
  private static String IMPORT_SOURCE = "source";
  private static String IMPORT_SOURCE_URL = "sourceUrl";
  private static String IMPORT_AUTHOR = "author";
  private static String IMPORT_DISPLAY_URL = "url";

  protected Member author;
  protected String authorDBID;
  protected Member opAuthor;
  protected Member opDelegate;
  protected DataExtension extension;
  protected HashMap<String, Object> importMap;
  protected HashMap<String, String> extraDataMap;

  transient private boolean isNewImport;
  transient private Long rowId;
  transient private String authorId;
  
  transient private long previousMDate;
  

  public Data() {
  }

  public Data(Member author, DataExtension extension) {
    this.author = author;
    this.extension = extension;
  }

  public Data(Data other) {
    super(other);
    this.author = other.author;
    if (!isDBData()) {
      this.extension = other.extension;
      this.extraDataMap = other.extraDataMap;
    }
    this.importMap = other.importMap;
    this.rowId = other.rowId;
  }

  // -------------------------------------------------------------------------
  // Data Name ("title", "name")
  // -------------------------------------------------------------------------

  /**
   * Return the name of this Data, every sub class Data which will be displayed to user should
   * implements this method already. <br>
   * <ul>
   * <li>The Title (multi-language) for a Publication.</li>
   * <li>The Name (multi-language) for a Category.</li>
   * <li>The Name for a Group, a Member. <br>
   * <li>The Title (multi-language) for a Workspace.</li>
   * </ul>
   * Can be seen as a smart multi-language toString(), but never returning null. <br>
   * Default behaviour is to return toFullString() of Data or empty string if toFullString is null.
   * 
   * @param language the language in which to return the name if multilang available for the data.
   * @return a String, MUST NOT return null.
   */
  public String getDataName(String language) {
    return Util.getString(toFullString(), "");
  }

  // -------------------------------------------------------------------------
  // LangPropertyArgument implementation
  // -------------------------------------------------------------------------

  /**
   * Retrieve the string value of this object in the specified language, 
   * when being used inside a language property.
   * <p>
   * This method is the implementation of {@link LangPropertyArgument#getLangPropertyValue(String)}
   * for automatic translation of Object in language properties.<br/>
   * You should not invoke this method directly, use {@link #getDataName(String)} instead.
   * 
   * @param language the ISO-639 language code to be used to retrieve the localized Object's name
   * @return this object as a string value in the specified language
   */
  public String getLangPropertyValue(String language) {
    return getDataName(language);
  }

  // -------------------------------------------------------------------------
  // Data Image ("thumbnail", "visuel", "illustration")
  // -------------------------------------------------------------------------

  /**
   * Retrieves the image path of this data in the channel's main language.
   * <p>
   * This method can (should) be implemented by subclasses to provide an image/illustration of this Data.<br>
   * Default behaviour is to return empty string.
   * 
   * @return a path relative to the webapp root.
   * @since jcms-6.0.0
   */
  public String getDataImage() {
    return "";
  }

  /**
   * Retrieves the image path of this data.
   * <p>
   * This method can (should) be implemented by subclasses to provide an image/illustration of this Data.<br>
   * Default behaviour is to return empty string.
   * 
   * @param lang the user language (ISO-639 language code)
   * @param useDefault should the default language be used if no value is found for the specified language
   * @return a path relative to the webapp root.
   * @since jcms-6.0.0
   */
  public String getDataImage(String lang, boolean useDefault) {
    return getDataImage();
  }

  /**
   * Retrieves the image path of this data in the specified language.
   * 
   * @param lang the user language (ISO-639 language code)
   * @return a path relative to the webapp root.
   * @since jcms-6.0.0
   */
  public String getDataImage(String lang) {
    return getDataImage(lang, true);
  }

  /**
   * Returns a generic thumbnail for this Data.
   * As defined by properties <code>types.{short-class-name}.thumb</code>.
   * <p>
   * If no property was defined for the class of this Data, the thumbnail specified in property <code>types.unknown.thumb</code> is used.
   * <p>
   * FileDocument uses its own implementation, see {@link FileDocument#getGenericThumbnail()}.
   * 
   * @return a thumbnail path relative to the webapp root, eg <code>images/jalios/icons/data/article.gif</code>
   * @since jcms-6.1.1
   */
  public String getGenericThumbnail() {
    String thumb = channel.getProperty(JcmsConstants.TYPES_PREFIX_PROP + Util.getClassShortName(this) + JcmsConstants.TYPES_THUMB_SUFFIX_PROP, null);
    if (thumb != null) {
      return thumb;
    }
    return channel.getProperty(JcmsConstants.TYPES_PREFIX_PROP + "unknown" + JcmsConstants.TYPES_THUMB_SUFFIX_PROP);
  }

  // -------------------------------------------------------------------------
  // toString
  // -------------------------------------------------------------------------

  @Override
  public String toString() {
    if (isDBData()) {
      return String.valueOf(getRowId());
    }
    return toFullString();
  }

  public String toFullString() {
    return "[class: " + getClass().getName() + "] " + "[id: " + id + "] " + "[author: " + (author == null ? "null" : author.getName()) + "] ";
  }

  // -------------------------------------------------------------------------
  // CSV
  // -------------------------------------------------------------------------

  /**
   * Returns a CSV view of this object
   *
   * @param lang the output language
   * @return a CSV view of this object
   * @since jcms-5.0.0
   */
  public String toCSV(String lang) {
    StringBuilder sb = new StringBuilder();
    // @formatter:off
    sb.append(JcmsUtil.convertToCSV(lang, cdate)            ).append(';')
      .append(JcmsUtil.convertToCSV(lang, mdate)            ).append(';')
      .append(JcmsUtil.convertToCSV(lang, author, false)    ).append(';')
      .append(JcmsUtil.convertToCSV(lang, author, true)     ).append(';')
      .append(JcmsUtil.convertToCSV(lang, id)               );
    // @formatter:on
    String csvLine = sb.toString();
    return csvLine;
  }

  /**
   * Returns the CSV header of this object
   *
   * @param lang the output language
   * @return the CSV header of this object
   * @since jcms-5.0.0
   */
  public static String getCSVHeader(String lang) {
    char quote = '"';
    return
      quote + JcmsUtil.glp(lang, "ui.com.lbl.cdate")  + quote + ";" +
      quote + JcmsUtil.glp(lang, "ui.com.lbl.mdate")  + quote + ";" +
      quote + JcmsUtil.glp(lang, "ui.com.lbl.author") + quote + ";" +
      quote + JcmsUtil.glp(lang, "ui.com.lbl.author")
           + '-' + JcmsUtil.glp(lang, "ui.com.lbl.id")+ quote + ";" +
      quote + JcmsUtil.glp(lang, "ui.com.lbl.id")     + quote;
  }

  /**
   * Update the given index.<br>
   * <br>
   * When used inside a StoreListener, in the method {@link com.jalios.jstore.StoreListener#handleCommitUpdate(com.jalios.jstore.Storable, com.jalios.jstore.Storable, boolean)},
   * the oldStorable can be given as a parameter of this method to remove it from the index,
   * Indeed removing old object from TreeSet or HashSet index works because both object
   * are the same when compared using {@link com.jalios.jstore.BasicStorable#compareTo(BasicStorable)} or {@link com.jalios.jstore.BasicStorable#hashCode()} methods.
   * 
   * @param indexName the name of the index
   * @param obj the object to be added or removed
   * @param add add if true, remove if false
   * @since jcms-5.0.1
   */
  protected synchronized void updateIndexTreeSet(String indexName, Object obj, boolean add) {
    try {
      Field field = getClass().getDeclaredField(indexName);

      TreeSet originalSet = (TreeSet) field.get(this);
      TreeSet newSet = originalSet;
      if (channel.isAvailable()) { // :NOTE:OD: only copy after channel startup (optimization)
        newSet = new TreeSet(originalSet);
      }
      if (add) {
        newSet.add(obj);
      } else {
        newSet.remove(obj);
      }

      if (originalSet != newSet) { // optimization
        field.set(this, newSet);
      }
    } catch (Exception ex) {
      logger.warn("An exception occured while " + (add ? "adding" : "removing") + " '" + obj + "' in index '" + indexName + "' for " + Util.getClassShortName(this) + " '" + this + "'", ex);
    }
  }

  /**
   * Returns the Set of clazz instances referring this data
   * 
   * @param <T> class of wanted objects
   * @param clazz the class to search the referring instances
   * @return a TreeSet of Data
   * @since jcms-3.0
   */
  public <T> TreeSet<T> getLinkIndexedDataSet(Class<T> clazz) {
    return channel.getLinkIndexedDataSet(this, clazz);
  }

  /**
   * A convenient method which returns the Set of clazz instances
   * referring this data with a given field. (it delegates on
   * channel.getLinkIndexedDataSet(this, clazz, field))
   * 
   * @param <T> class of wanted objects
   * @param clazz the class to search the referring instances
   * @param field the name of the field which contains the reference
   * @return TreeSet
   * @since jcms-4.1
   */
  public <T> TreeSet<T> getLinkIndexedDataSet(Class<T> clazz, String field) {
    return channel.getLinkIndexedDataSet(this, clazz, field);
  }

  /**
   * @return the number of instances referring this data
   * @since jcms-3.0
   */
  public int getLinkCount() {
    return channel.getLinkCount(this);
  }

  
  /**
   * Returns the referrer set of this data.
   * @return the referrer set of this data.
   * @since jcms-9.0.0
   */
  public Set getReferrerSet() {
    Map<Data, Map<Class<?>, Map<String, TreeSet<Data> > > > index = channel.getLinkIndexManager().getIndex();
    Set<Data> referrerSet = new HashSet<Data>();
    Map<Class<?>, Map<String, TreeSet<Data>>> referrerMap = index.get(this);
    if (referrerMap != null) {
      for(Class itClass: referrerMap.keySet()) {
        for(Object referrer: getLinkIndexedDataSet(itClass)) {
          referrerSet.add((Data)referrer);
        }
      }
    }
    
    return referrerSet;
  }
  
  /**
   * Return all the referrers of this data, instance of the given class
   * 
   * @param referrerClass the class to search the referring instances
   * @return TreeSet
   * @since jcms-4.1
   */
  public TreeSet getAllReferrerSet(Class referrerClass) {
    TreeSet resultSet = new TreeSet();
    if (referrerClass == null) {
      return resultSet;
    }
    TreeSet visitedSet = new TreeSet();
    getAllReferrerSetVisitor(this, referrerClass, resultSet, visitedSet);
    return resultSet;
  }

  /**
   * Return all the referrers of this data
   * 
   * @return TreeSet
   * @since jcms-4.1
   */
  public TreeSet getAllReferrerSet() {
    return getAllReferrerSet(Data.class);
  }

  /**
   * @since jcms-4.1
   */
  void getAllReferrerSetVisitor(Data data, Class<Data> referrerClass, TreeSet<Data> resultSet, TreeSet<Data> visitedSet) {
    visitedSet.add(data);

    // Iterates over referrers
    TreeSet<Data> referrerSet = data.getLinkIndexedDataSet(Data.class);
    for (Iterator<Data> it = referrerSet.iterator(); it.hasNext();) {
      Data referrer = it.next();

      // Visits referrers not yet visited
      if (!visitedSet.contains(referrer)) {
        if (referrerClass.isAssignableFrom(referrer.getClass())) {
          resultSet.add(referrer);
        }
        getAllReferrerSetVisitor(referrer, referrerClass, resultSet, visitedSet);
      }
    }
  }

  /**
   * Fill the given string buffer with specific attributes of this data.
   * 
   * @param sb the StringBuffer
   * @since jcms-5.7.1
   */
  public void fillExportXmlFieldAttributes(StringBuffer sb) {
    sb.append(" id='").append(getId()).append("'");
    sb.append(" class='").append(getClass().getName()).append("'");
  }

  /**
   * Returns a string representation of this data used by exportXml().
   * 
   * @return a string representation of this data used by exportXml().
   * @since jcms-5.7.2
   */
  public String getExportXmlDataValue() {
    return XmlUtil.normalize(getDataName(channel.getLanguage()));
  }

  /**
   * Fill this data with the values contains in the JDOM element.
   * 
   * @param elt the JDOM element
   * @param options the ImportOptions
   * @since jcms-5.7.0
   */
  public void importXml(Element elt, ImportOptions options) {
    try {
      
      if (options.isSelfImport()) {
        
        setAuthor(ImportUtil.parseSelfFieldData(elt, "author", Member.class));
        
      } else {
        
        setImportId(elt.getAttributeValue("id"));
        setImportMdate(DateUtil.parseW3cDate(ImportUtil.parseFieldText(elt, "mdate")));

        setImportDisplayUrl(elt.getAttributeValue("url"));
        setImportAuthor(ImportUtil.parseFieldMemberLogin(elt, "author"));

        setAuthor(options.getImportAuthor());
        
      }

      importXmlFieldsWithReferences(elt, options);
    } catch (Exception ex) {
      logger.error("An error occured while importing XML " + elt, ex);
    }
  }

  /**
   * Resolves imported references for this data.
   * 
   * @param elt the JDOM element
   * @param options the ImportOptions
   * @since jcms-6.0.0
   */
  public void importXmlReferences(Element elt, ImportOptions options) {
    importXmlFieldsWithReferences(elt, options);
  }

  protected void importXmlFieldsWithReferences(Element elt, ImportOptions options) {
    // EMPTY. It will overrided in sub-classes
  }

  public void importXmlFiles(Map fieldFileMap, ImportOptions options) {
    // EMPTY. It will overrided in sub-classes
  }

  /**
   * Returns the XML export this data.
   * 
   * @return the XML export for the given data. The returned String does not contain the XML declaration (<?xml>).
   * @since jcms-5.7.2
   */
  public String exportXml() {
    return exportXml(null);
  }

  /**
   * Returns the XML export this data.
   * 
   * @param relateds possible relateds to be added to the export
   * @return the XML export for the given data. The returned String does not contain the XML declaration (<?xml>).
   * @since jcms-5.7.2
   */
  public String exportXml(String[] relateds) {
    StringBuffer sb = new StringBuffer(1024);
    return exportXml(sb, relateds).toString();
  }

  /**
   * Fill the given StringBuffer with the XML export this data. The StringBuffer does not contain the XML declaration (<?xml>).
   * 
   * @param sb the StringBuffer
   * @param relateds possible relateds to be added to the export
   * @return the given in parameter sb parameter with content added
   * @since jcms-5.7.2
   */
  public StringBuffer exportXml(StringBuffer sb, String[] relateds) {
    exportXml(sb, relateds, 0);
    return sb;
  }

  /**
   * Fill the given StringBuffer with the XML export this data. The StringBuffer does not contain the XML declaration (<?xml>).
   * 
   * @param sb the StringBuffer.
   * @param relateds possible relateds to be added to the export
   * @param indentLevel indentation level to be used.
   * @return the given in parameter sb parameter with content added
   * @since jcms-5.7.2
   */
  public StringBuffer exportXml(StringBuffer sb, String[] relateds, int indentLevel) {
    String indent = ExportUtil.getIndent(indentLevel);
    String url = channel.getUrl() + getDisplayUrl(channel.getLocale());

    // Open tag
    sb.append(indent).append("<" + DATA_TAG + " class='").append(getClass().getName()).append("' id='").append(id).append("' url='").append(url).append("'>\n");

    // Field tags
    exportXmlField(sb, indentLevel + 1);
    if (Util.notEmpty(relateds)) {
      exportXmlRelated(sb, relateds, indentLevel + 1);
    }

    // Close tag
    sb.append(indent + "</" + DATA_TAG + ">\n");
    return sb;
  }

  /**
   * Fills the given StringBuffer with Xml export of fields of this data.
   * 
   * @param sb the StringBuffer
   * @param indentLevel the indent level
   * @since jcms-5.7.2
   */
  public void exportXmlField(StringBuffer sb, int indentLevel) {
    sb
    .append(ExportUtil.exportField(indentLevel, "author", getAuthor()))
    .append(ExportUtil.exportField(indentLevel, "opAuthor", opAuthor))
    .append(ExportUtil.exportField(indentLevel, "cdate", cdate))
    .append(ExportUtil.exportField(indentLevel, "mdate", mdate));
  }

  /**
   * Fills the given StringBuffer with Xml export of related of this data.
   * Related are data computed or stored in memory (index), but not persisted as is.
   * Only invoked related are given (through relateds parameter).
   * 
   * @param sb the StringBuffer
   * @param relateds possible relateds to be added to the export
   * @param indentLevel the indent level
   * @since JCMS-6.1
   */
  public void exportXmlRelated(StringBuffer sb, String[] relateds, int indentLevel) {
    for (int i = 0; i < relateds.length; i++) {
      String related = relateds[i];
      if (Util.notEmpty(related)) {
        exportXmlRelated(sb, related, indentLevel);
      }
    }
  }

  @SuppressWarnings("unchecked")
  protected void exportXmlRelated(StringBuffer sb, String related, int indentLevel) {
  	channel.getPolicyManager().processDataXmlRelatedExport(this, sb, related);
  	
  	if (related.startsWith("extraData.")) {
  		String name = related.substring("extraData.".length());
  		sb.append(ExportUtil.exportRelated(indentLevel, related, getExtraData(name)));
  	} else if (related.startsWith("extraDBData.")) {
  		String name = related.substring("extraDBData.".length());
  		sb.append(ExportUtil.exportRelated(indentLevel, related, getExtraDBData(name)));
    } else if ("allExtraData".equals(related)) {
  		getRelatedAllExtraData(sb, indentLevel);
  	} else if (related.startsWith("extraInfo.")) {
      String name = related.substring("extraInfo.".length());
      sb.append(ExportUtil.exportRelated(indentLevel, related, getExtraInfo(name)));
    } else if (related.equals("dataImage")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getDataImage()));
    } else if (related.equals("linkCount")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getLinkCount()));
    } else if (related.equals("isLocked")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, isLocked()));
    } else if (related.equals("lockMember")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getLockMember()));
    } else if (related.equals("lockDate")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getLockDate()));
    } else if (related.equals("isStrongLocked")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, isStrongLocked()));
    } else if (related.equals("strongLockMember") && Util.notEmpty(getStrongLock())) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getStrongLock().getAuthor()));
    } else if (related.equals("strongLockDate") && Util.notEmpty(getStrongLock())) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getStrongLock().getCdate()));
    } else if (related.equals("virtualID")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getVirtualIDSet()));
    } else if (related.equals("allReferrerSet")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getAllReferrerSet()));
    } else if (related.startsWith("allReferrerSet.")) {
      String className = related.substring("allReferrerSet.".length());
      try {
        sb.append(ExportUtil.exportRelated(indentLevel, related, getAllReferrerSet(channel.getClass(className))));
      } catch (ClassNotFoundException ex) {
      }
    } else if (related.startsWith("linkIndexedDataSet.Class.")) {
      String className = related.substring("linkIndexedDataSet.Class.".length());
      try {
        sb.append(ExportUtil.exportRelated(indentLevel, related, getLinkIndexedDataSet(channel.getClass(className))));
      } catch (ClassNotFoundException ex) {
      }
    } else if (related.startsWith("linkIndexedDataSet.ClassField.")) {
      String classAndFieldName = related.substring("linkIndexedDataSet.ClassField.".length());
      String fieldName = classAndFieldName.substring(classAndFieldName.lastIndexOf('.'));
      String className = classAndFieldName.substring(0, classAndFieldName.lastIndexOf('.'));
      try {
        sb.append(ExportUtil.exportRelated(indentLevel, related, getLinkIndexedDataSet(channel.getClass(className), fieldName)));
      } catch (ClassNotFoundException ex) {
      }
    } else if (related.equals("lockMember")) {
      sb.append(ExportUtil.exportRelated(indentLevel, related, getLockMember()));
    }
  }

  /**
   * Fills the given StringBuffer with Xml export of all related extra data of this data.
   * @param sb the StringBuffer
   * @param indentLevel the indent level
   */
	private void getRelatedAllExtraData(StringBuffer sb, int indentLevel) {
		Map<String, String>	extraDBData = getExtraDBDataMap();
		Map<String, String>	extraDataMap = null;
		
		if(!isDBData()){
			extraDataMap = getExtraDataMap();
		}
		
		if(extraDataMap != null){
			if(extraDBData != null){
				extraDataMap.putAll(extraDBData);
			}
		}else{
			if(extraDBData != null){
				extraDataMap = new HashMap<String, String>(extraDBData);
			}
		}
		
		if(extraDataMap == null || Util.isEmpty(extraDataMap)) return;
		
		Set<Entry<String, String>> entrySet =	extraDataMap.entrySet();
		for (Entry<String, String> entry : entrySet) {
			sb.append(ExportUtil.exportRelated(indentLevel, entry.getKey(), entry.getValue()));
		}
	}

  /**
   * Returns true if this data has been imported.
   * 
   * @return true if this data has been imported.
   * @since jcms-5.7.2
   */
  public boolean isImported() {
    return Util.notEmpty(getImportId());
  }

  /**
   * Set the value of mdate
   * 
   * @param v Value to assign to mdate.
   * @since jcms-5.7.2
   */
  @Override
  public void setMdate(Date v) {
    super.setMdate(v);

    // Assign the importDate
    if (isImported() && isNewImport) {
      setImportDate(v);
      isNewImport = false;
    }
  }

  /**
   * Internal method used to mark an imported data. Use with care.
   * 
   * @since jcms-5.7.2
   */
  public void markNewImport() {
    isNewImport = true;
  }

  /**
   * Internal method used to unmark an imported data. Use with care.
   * 
   * @since jcms-6.0.0
   */
  public void unmarkNewImport() {
    isNewImport = false;
  }

  /**
   * Get the value of importMap.
   * 
   * @return Value of importMap.
   * @since jcms-5.7.2
   */
  public HashMap<String, Object> getImportMap() {
    return importMap;
  }

  /**
   * Set the value of importMap.
   * 
   * @param v Value to assign to importMap.
   * @since jcms-5.7.2
   */
  public void setImportMap(HashMap<String, Object> v) {
    importMap = v;
  }

  /**
   * Get the value of importId.
   * 
   * @return Value of importId.
   * @since jcms-5.7.2
   */
  public String getImportId() {
    if (importMap == null) {
      return null;
    }
    return (String) importMap.get(IMPORT_ID);
  }

  /**
   * Set the value of importId.
   * 
   * @param v Value to assign to importId.
   * @since jcms-5.7.2
   */
  public void setImportId(String v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_ID, v);
  }

  /**
   * Get the value of importDate.
   * 
   * @return Value of importDate.
   * @since jcms-5.7.2
   */
  public Date getImportDate() {
    if (importMap == null) {
      return null;
    }
    return (Date) importMap.get(IMPORT_DATE);
  }

  /**
   * Set the value of importDate.
   * 
   * @param v Value to assign to importDate.
   * @since jcms-5.7.2
   */
  public void setImportDate(Date v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_DATE, v);
  }

  /**
   * Get the value of importMdate.
   * 
   * @return Value of importMdate.
   * @since jcms-5.7.2
   */
  public Date getImportMdate() {
    if (importMap == null) {
      return null;
    }
    return (Date) importMap.get(IMPORT_MDATE);
  }

  /**
   * Set the value of importMdate.
   * 
   * @param v Value to assign to importMdate.
   * @since jcms-5.7.2
   */
  public void setImportMdate(Date v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_MDATE, v);
  }

  /**
   * Get the value of importBatch.
   * 
   * @return Value of importBatch.
   * @since jcms-5.7.2
   */
  public String getImportBatch() {
    if (importMap == null) {
      return null;
    }
    return (String) importMap.get(IMPORT_BATCH);
  }

  /**
   * Set the value of importBatch.
   * 
   * @param v Value to assign to importBatch.
   * @since jcms-5.7.2
   */
  public void setImportBatch(String v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_BATCH, v);
  }

  /**
   * Get the value of isImportUpdated.
   * 
   * @return Value of isImportUpdated.
   * @since jcms-5.7.2
   */
  public boolean isImportUpdated() {
    if (importMap == null) {
      return false;
    }
    return importMap.get(IMPORT_UPDATED) == Boolean.TRUE;
  }

  /**
   * Set the value of isImportUpdated.
   * 
   * @param v Value to assign to isImportUpdated.
   * @since jcms-5.7.2
   */
  public void setImportUpdated(boolean v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_UPDATED, Boolean.valueOf(v));
  }

  /**
   * Get the value of importSignature.
   * 
   * @return Value of importSignature.
   * @since jcms-5.7.2
   */
  public String getImportSignature() {
    if (importMap == null) {
      return null;
    }
    return (String) importMap.get(IMPORT_SIGNATURE);
  }

  /**
   * Set the value of importSignature.
   * 
   * @param v Value to assign to importSignature.
   * @since jcms-5.7.2
   */
  public void setImportSignature(String v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_SIGNATURE, v);
  }

  /**
   * Get the value of importSource.
   * 
   * @return Value of importSource.
   * @since jcms-5.7.2
   */
  public String getImportSource() {
    if (importMap == null) {
      return null;
    }
    return (String) importMap.get(IMPORT_SOURCE);
  }

  /**
   * Set the value of importSource.
   * 
   * @param v Value to assign to importSource.
   * @since jcms-5.7.2
   */
  public void setImportSource(String v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_SOURCE, v);
  }

  /**
   * Get the value of importSourceUrl.
   * 
   * @return Value of importSourceUrl.
   * @since jcms-5.7.2
   */
  public String getImportSourceUrl() {
    if (importMap == null) {
      return null;
    }
    return (String) importMap.get(IMPORT_SOURCE_URL);
  }

  /**
   * Set the value of importSourceUrl.
   * 
   * @param v Value to assign to importSourceUrl.
   * @since jcms-5.7.2
   */
  public void setImportSourceUrl(String v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_SOURCE_URL, v);
  }

  /**
   * Get the value of importAuthor.
   * 
   * @return Value of importAuthor.
   * @since jcms-5.7.2
   */
  public String getImportAuthor() {
    if (importMap == null) {
      return null;
    }
    return (String) importMap.get(IMPORT_AUTHOR);
  }

  /**
   * Set the value of importAuthor.
   * 
   * @param v Value to assign to importAuthor.
   * @since jcms-5.7.2
   */
  public void setImportAuthor(String v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_AUTHOR, v);
  }

  /**
   * Get the value of importDisplayUrl.
   * 
   * @return Value of importDisplayUrl.
   * @since jcms-5.7.2
   */
  public String getImportDisplayUrl() {
    if (importMap == null) {
      return null;
    }
    return (String) importMap.get(IMPORT_DISPLAY_URL);
  }

  /**
   * Set the value of importAuthor.
   * 
   * @param v Value to assign to importAuthor.
   * @since jcms-5.7.2
   */
  public void setImportDisplayUrl(String v) {
    if (importMap == null) {
      importMap = new HashMap<String, Object>();
    }
    importMap.put(IMPORT_DISPLAY_URL, v);
  }

  /**
   * Returns the import URL of this publication, i.e. the XML export URL of the original data.
   * 
   * @return the import URL of this publication.
   * @since jcms-5.7.2
   */
  public String getImportUrl() {
    String importId = getImportId();

    if (Util.isEmpty(importId)) {
      return null;
    }

    String importSourceUrl = getImportSourceUrl();
    if (Util.isEmpty(importSourceUrl)) {
      return null;
    }

    return importSourceUrl + "admin/exportXml.jsp?id=" + importId;
  }

  /**
   * Returns the author of this Data. If the author is a DBMember it is loaded.
   * This method must not be called before the database as been initialized (e.g. it must not be called by a StoreListener).
   * The author returned by method must not be kept in data structures that are longer than the DB transaction (e.g. an index).
   * In those cases, call getAuthor(false).
   * 
   * @return the author of this Data.
   * @see #getAuthor(boolean)
   */
  public Member getAuthor() {
    return getAuthor(true);
  }

  /**
   * Returns the author of this Data. If queryDB is true query the DB to find the author from authorDBID.
   * 
   * @param queryDB indicates if the database must be re.
   * @return the author of this Data.
   * @since jcms-6.1.0
   */
  public Member getAuthor(boolean queryDB) {
    if (author != null) {
      return author;
    }
    if (queryDB && Util.notEmpty(authorDBID)) {
      return channel.getMember(authorDBID);
    }
    return null;
  }

  /**
   * Sets the author of this Data
   * 
   * @param v the author
   * @since jcms-6.1.0
   */
  public void setAuthor(Member v) {
    if (v instanceof DBData) {
      setAuthorDBID(v.getId());
    } else {
      author = v;
      authorId = null;
    }
  }

  /**
   * Returns a JCMS ID of the author if it is persisted in the DB.
   * 
   * @return a JCMS ID.
   * @since jcms-6.1.0
   */
  public String getAuthorDBID() {
    return authorDBID;
  }

  /**
   * Sets the authorDBID of this Data.
   * 
   * @param v the authorDBID
   * @since jcms-6.1.0
   */
  public void setAuthorDBID(String v) {
    authorDBID = v;
    if (Util.notEmpty(authorDBID)) {
      author = null;
    }
  }

  /**
   * Get the value of opAuthor.
   * 
   * @return Value of opAuthor.
   * @since jcms-4.0
   */
  public Member getOpAuthor() {

    if (opAuthor == null && this instanceof HistorizedDBData) {
      DBDataRevision rev = Util.getLast(DBDataRevisionManager.getInstance().getRevisionList(this));
      if (rev != null) {
        return rev.getRevAuthor();
      }
    }
    
    return opAuthor;
  }

  /**
   * Set the value of opAuthor.
   * 
   * @param v Value to assign to opAuthor.
   * @since jcms-4.0
   */
  public void setOpAuthor(Member v) {
    this.opAuthor = v;
  }
  
  /**
   * Retrieves the Member who is really doing the operation
   * (through right delegation) as opposed to the opAuthor,
   * who is the Member having the right to do it.
   * 
   * @return Value of opDelegate.
   * @since jcms-5.7.0
   */
  public Member getOpDelegate() {
    return opDelegate;
  }

  /**
   * Defines the Member who is really doing the operation
   * (through right delegation) as opposed to the opAuthor,
   * who is the Member having the right to do it.
   * 
   * @param mbr Value to assign to opDelegate.
   * @since jcms-5.7.0
   */
  public void setOpDelegate(Member mbr) {
    this.opDelegate = mbr;
  }

  /**
   * Get the value of extension.
   * 
   * @return Value of extension.
   */
  public DataExtension getExtension() {
    return extension;
  }

  /**
   * Set the value of extension.
   * 
   * @param v Value to assign to extension.
   */
  public void setExtension(DataExtension v) {
    this.extension = v;
    if (extension != null) {
      extension.setExtendedData(this);
    }
  }

  // ----------------------------------------------------------------------
  // Date getter
  // ----------------------------------------------------------------------

  /**
   * Return the date corresponding to the given parameter :
   * date/cdate, mdate or null if nothing match or the matching date
   * is null.
   * 
   * @param dateType the date type
   * @return the date
   * @since jcms-4.0
   */
  public Date getDate(String dateType) {
    return getDate(dateType, null);
  }

  /**
   * Return the date corresponding to the given parameter :
   * date/cdate, mdate, or null if nothing match or the matching date
   * is null.
   * 
   * @param dateType the date type
   * @param defaultDate the default date
   * @return the date
   * @since jcms-4.0
   */
  public Date getDate(String dateType, Date defaultDate) {
    if (Util.isEmpty(dateType)) {
      return defaultDate;
    }

    Date date = null;

    if (dateType.equalsIgnoreCase("mdate")) {
      date = getMdate();
    }

    if (dateType.equalsIgnoreCase("date") || dateType.equalsIgnoreCase("cdate")) {
      date = getCdate();
    }

    if (date != null) {
      return date;
    }

    return defaultDate;
  }

  // ----------------------------------------------------------------------
  // Selector getter
  // ----------------------------------------------------------------------

  public static CdateSelector getCdateSelector(Date begin, Date end) {
    return new CdateSelector(begin, end);
  }

  public static MdateSelector getMdateSelector(Date begin, Date end) {
    return new MdateSelector(begin, end);
  }

  /**
   * Retrieves a new {@link DataSelector} given a name representing
   * it. The returned selector can be passed to the method {@link JcmsUtil#applyDataSelector(java.util.Collection, DataSelector)} to
   * allow precise control over the selection of <code>Data</code> in a <code>Collection</code>. <br>
   * <br>
   * You can specify any of the following value: <li><code>null</code>, <tt>"date"</tt>, <tt>"cdate"</tt>: {@link #getCdateSelector(Date begin, Date end)}</li> <li><tt>"mdate"</tt>: {@link #getMdateSelector(Date begin, Date end)}</li>
   * 
   * @param str a string representing the selector
   * @param begin the begin <code>Date</code> to use for selection
   * @param end the end <code>Date</code> to use for selection
   * @return a new instance of <code>Comparator</code>.
   */
  public static DataSelector getDateSelector(String str, Date begin, Date end) {
    DataSelector selector = null;

    if (str == null || str.equalsIgnoreCase("date") || str.equalsIgnoreCase("cdate")) {
      selector = getCdateSelector(begin, end);
    } else if (str.equalsIgnoreCase("mdate")) {
      selector = getMdateSelector(begin, end);
    }

    return selector;
  }

  public static DataSelector getImportSelector() {
    return new ImportSelector();
  }

  public static DataSelector getDeletableSelector(Member opAuthor) {
    return new DeletableSelector(opAuthor);
  }

  public static DataSelector getAuthorSelector(Member author) {
    return new AuthorSelector(author);
  }

  // ----------------------------------------------------------------------
  // Comparator getter
  // ----------------------------------------------------------------------
  /**
   * @param <T> subclass of Data
   * @return a new instance of {@link CaddyComparator}.
   * @since jcms-5.7
   */
  public static <T extends Data> Comparator<T> getCaddyComparator() {
    return new CaddyComparator<T>();
  }

  /**
   * @param <T> subclass of Data
   * @param memberComparator a member comparator
   * @return a new instance of {@link AuthorComparator} built with the given
   * member comparator, usually a {@link Member#getNameComparator()}.
   */
  public static <T extends Data> Comparator<T> getAuthorComparator(Comparator<Member> memberComparator) {
    return new AuthorComparator<T>(memberComparator);
  }

  /**
   * @return a new instance of {@link OpAuthorComparator} built with the given
   * member comparator, usually a {@link Member#getNameComparator()}.
   * @param <T> subclass of Data
   * @param memberComparator a Member comparator
   * @since jcms-4.0
   */
  public static <T extends Data> Comparator<T> getOpAuthorComparator(Comparator<Member> memberComparator) {
    return new OpAuthorComparator<T>(memberComparator);
  }

  /**
   * @param <T> subclass of Data
   * @return a new DataNameComparator built with the given language.
   * @param language the language that will be used by the DataNameComparator
   * @since jcms-5.5.0
   */
  public static <T extends Data> Comparator<T> getDataNameComparator(String language) {
    return new DataNameComparator<T>(language);
  }

  /**
   * @param <T> subclass of Data
   * @return a new RowIdComparator
   * @since jcms-6.2.0
   */
  public static <T extends Data> Comparator<T> getRowIdComparator() {
    return new RowIdComparator<T>();
  }

  /**
   * Retrieves a new <code>Comparator</code> given a name representing
   * it. The returned comparator can be used to sort any <code>Collection</code> containing <code>Data</code>s. <br>
   * You can specify any of the following value: <li><code>null</code>, <tt>"date"</tt>, <tt>"cdate"</tt>: {@link #getCdateComparator()}</li> <li><tt>"mdate"</tt>: {@link #getMdateComparator()}</li> <li><tt>"author"</tt>:
   * {@link #getAuthorComparator(Comparator)} with a {@link Member#getNameComparator()}</li> <li><tt>"opauthor"</tt>: {@link #getOpAuthorComparator(Comparator)} with a {@link Member#getNameComparator()}</li> <li><tt>"id"</tt>:
   * {@link #getIdComparator()}</li> <li><tt>"caddy"</tt>: {@link #getCaddyComparator()}, since jcms-5.7.</li>
   * 
   * @param <T> specialization of Data
   * @param str a string representing the comparator
   * @param reverse a boolean indicating whether the comparator should have
   * its behavior reversed (<tt>true</tt>) or unchanged (<tt>false</tt>)
   * @return a new instance of <code>Comparator</code>.
   * @deprecated use {@link ComparatorManager}
   */
  public static <T extends Data> Comparator<T> getComparator(String str, boolean reverse) {
    Comparator<T> comparator = null;

    if (Util.isEmpty(str) || str.equalsIgnoreCase("date") || str.equalsIgnoreCase("cdate")) {
      comparator = getCdateComparator();
    } else if (str.equalsIgnoreCase("mdate")) {
      comparator = getMdateComparator();
    } else if (str.equalsIgnoreCase("author")) {
      comparator = getAuthorComparator(Member.getNameComparator());
    } else if (str.equalsIgnoreCase("opauthor")) {
      comparator = getOpAuthorComparator(Member.getNameComparator());
    } else if (str.equalsIgnoreCase("id")) {
      comparator = getIdComparator();
    } else if (str.equalsIgnoreCase("caddy")) {
      comparator = getCaddyComparator();
    } else if (str.equalsIgnoreCase("rowId")) {
      comparator = getRowIdComparator();
    }

    if (reverse && comparator != null) {
      comparator = new ReverseComparator<T>(comparator);
    }

    return comparator;
  }

  // ----------------------------------------------------------------------
  // Selector classes
  // ----------------------------------------------------------------------

  /**
   * <code>DataSelector</code> to select <code>Data</code> which have
   * a creation date between the given begin and end <code>Date</code>.
   * 
   * @see Data#getCdate()
   * @see Data#getCdateSelector(Date, Date)
   */
  public static class CdateSelector implements DataSelector {
    Date begin;
    Date end;

    public CdateSelector(Date begin, Date end) {
      this.begin = begin;
      this.end = end;
    }

    public boolean isSelected(Data data) {

      Date cdate = data.getCdate();
      if (cdate == null) {
        return false;
      }
      return !cdate.before(begin) && !cdate.after(end);
    }
  }

  /**
   * <code>DataSelector</code> to select <code>Data</code> which have
   * a modification date between the given begin and end <code>Date</code>.
   * 
   * @see Data#getMdate()
   * @see Data#getMdateSelector(Date, Date)
   */
  public static class MdateSelector implements DataSelector {
    Date begin;
    Date end;

    public MdateSelector(Date begin, Date end) {
      this.begin = begin;
      this.end = end;
    }

    public boolean isSelected(Data data) {

      Date mdate = data.getMdate();
      if (mdate == null) {
        return false;
      }
      return !mdate.before(begin) && !mdate.after(end);
    }
  }

  /**
   * <code>DataSelector</code> to select <code>Data</code> which have
   * been imported.
   * 
   * @since jcms-5.7.2
   */
  public static class ImportSelector implements DataSelector {
    public boolean isSelected(Data data) {
      return data.isImported();
    }
  }

  /**
   * <code>DataSelector</code> to select <code>Data</code> that
   * can be deleted (i.e. data whose <code>checkDelete()</code> returns ok).
   * 
   * @since jcms-5.7.2
   */
  public static class DeletableSelector implements DataSelector {

    Member opAuthor = null;

    public DeletableSelector(Member opAuthor) {
      this.opAuthor = opAuthor;
    }

    public boolean isSelected(Data data) {
      return data.checkDelete(opAuthor).isOK();
    }
  }

  public static class AuthorSelector implements DataSelector {
    Member author = null;

    public AuthorSelector(Member author) {
      this.author = author;
    }

    public boolean isSelected(Data data) {
      return data.getAuthor() == author;
    }
  }

  /**
   * Retrieves a new instance of DataSelector based on the strong lock status of the
   * data.
   * 
   * @return a new instance of {@link StrongLockSelector}
   * @since jcms-6.2.0
   */
  public static DataSelector getStrongLockSelector() {
    return new StrongLockSelector();
  }

  /**
   * {@link DataSelector} selecting only {@link Data} having a
   * strong lock ({@link Data#isStrongLocked()}).
   * 
   * @since jcms-6.2.0
   */
  public static class StrongLockSelector implements DataSelector {

    public boolean isSelected(Data data) {
      return data.isStrongLocked();
    }
  }

  // ----------------------------------------------------------------------
  // Comparator classes
  // ----------------------------------------------------------------------

  public static class AuthorComparator<T extends Data> extends CdateComparator<T> {
    Comparator<Member> memberComparator;

    /**
     * Build a new Author comparator using a NameComparator to compare Member
     * @since jcms-7.1.0
     */
    public AuthorComparator() {
      this(Member.getNameComparator());
    }
    
    public AuthorComparator(Comparator<Member> memberComparator) {
      this.memberComparator = memberComparator;
    }

    @Override
    public int compare(T o1, T o2) {
      Member m1 = o1.getAuthor();
      Member m2 = o2.getAuthor();
      int res = memberComparator.compare(m1, m2);
      if (res != 0) {
        return res;
      }
      return super.compare(o1, o2);
    }
  }

  /**
   * @param <T> any Data type
   * @since jcms-4.0
   */
  public static class OpAuthorComparator<T extends Data> extends CdateComparator<T> {
    Comparator<Member> memberComparator;
    
    /**
     * Build a new OpAuthor comparator using a NameComparator to compare Member
     * @since jcms-7.1.0
     */
    public OpAuthorComparator() {
      this(Member.getNameComparator());
    }
    
    public OpAuthorComparator(Comparator<Member> memberComparator) {
      this.memberComparator = memberComparator;
    }

    @Override
    public int compare(T d1, T d2) {
      Member m1 = d1.getOpAuthor();
      Member m2 = d2.getOpAuthor();
      int res = memberComparator.compare(m1, m2);
      if (res != 0) {
        return res;
      }
      return super.compare(d1, d2);
    }
  }

  /**
   * This comparator compares two Data using their Data Name.
   * If both compared Data have the same Data Name, compare them using their id.
   * 
   * @param <T> any Data type
   * @since jcms-5.5.0
   */
  public static class DataNameComparator<T extends Data> extends BasicComparator<T> {

    protected String lang;

    public DataNameComparator() {
      this.lang = language;
    }
    
    public DataNameComparator(String language) {
      this.lang = language;
    }

    @Override
    public int compare(T storable1, T storable2) {
      String s1 = storable1.getDataName(lang);
      String s2 = storable2.getDataName(lang);

      if (s1 == null) {
        if (s2 == null) {
          return 0;
        }
        return -1;
      }
      if (s2 == null) {
        return 1;
      }

      int res = s1.compareToIgnoreCase(s2);
      if (res != 0) {
        return res;
      }
      return super.compare(storable1, storable2);
    }
  }

  public static class RowIdComparator<T extends Data> implements Comparator<T> {

    public int compare(T data1, T data2) {
      Long rowId1 = data1.getRowId();
      Long rowId2 = data2.getRowId();

      if (rowId1 == null) {
        if (rowId2 == null) {
          return 0;
        }
        return -1;
      }
      if (rowId2 == null) {
        return 1;
      }

      int res = rowId1.compareTo(rowId2);
      return res;
    }
  }

  // ----------------------------------------------------------------------
  // Extra Info Map
  // ----------------------------------------------------------------------

  /**
   * Stores an extra information in this <code>Data</code>. <br>
   * If the object passed in is null, the effect is the same as
   * calling {@link #removeExtraInfo}.<br>
   * <b>Data's extra information are transient (not stored upon a site restart)</b>.
   * 
   * @param name a <code>String</code> specifying the name of the extra info
   * @param obj the <code>Object</code> to be stored
   * @return previous value associated with specified name, or <tt>null</tt> if there was no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @since jcms-5.5.0
   */
  public Object setExtraInfo(String name, Object obj) {
    if (obj != null) {
      return DataIndexManager.getInstance().putExtraInfo(this, name, obj);
    }
    return removeExtraInfo(name);
  }

  /**
   * Removes an extra information from this <code>Data</code>.
   * 
   * @param name a <code>String</code> specifying the name of the extra info
   * @return previous value associated with specified name, or <tt>null</tt> if there was no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @since jcms-5.5.0
   */
  public Object removeExtraInfo(String name) {
    return DataIndexManager.getInstance().removeExtraInfo(this, name);
  }

  /**
   * Returns the value of the named extra info as an <code>Object</code>,
   * or <code>null</code> if no extra info of the given name exists.
   * 
   * @param name the name of the extrainfo to retrieve
   * @return previous value associated with specified name, or <tt>null</tt> if there is no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @since jcms-5.5.0
   */
  public Object getExtraInfo(String name) {
    return DataIndexManager.getInstance().getExtraInfo(this, name);
  }

  /**
   * Returns a java.util.Map of the extra infos of this results.
   * 
   * @return the <b>internal</b> Map used by this results, modify with care.
   * @since jcms-5.5.0
   */
  public Map<String, Object> getExtraInfoMap() {
    return DataIndexManager.getInstance().getExtraInfoMap(this);
  }

  // ----------------------------------------------------------------------
  // Stored State
  // ----------------------------------------------------------------------

  /**
   * Describes the current state of this data in its persistent storage system.
   */
  private enum StoredState {
    /** * New DBData not yet stored in database (hibernate) */
    DBDATA_NEW,
    /** * DBData stored in database (hibernate) */
    DBDATA_STORED,
    /** * New JStore Data not yet stored in store.xml */
    DATA_NEW,
    /** * JStore Data stored in store.xml, original instance */
    DATA_ORIGINAL,
    /** * JStore Data stored in store.xml, cloned instance */
    DATA_CLONED
  }

  /**
   * Retrieves the state of this data in its persistent storage system.
   * 
   * @return the StoredState of this data
   */
  private StoredState getStoredState() {
    // DBData
    if (isDBData()) {
      if (!isInDatabase()) {
        return StoredState.DBDATA_NEW;
      }
      return StoredState.DBDATA_STORED;
    }

    // JStore Data
    if (!isStored()) {
      return StoredState.DATA_NEW;
    }
    boolean isOriginal = (channel.getData(this.getId()) == this);
    if (isOriginal) {
      return StoredState.DATA_ORIGINAL;
    }
    return StoredState.DATA_CLONED;
  }

  // ----------------------------------------------------------------------
  // Extra Data Map
  // ----------------------------------------------------------------------

  /**
   * Returns a java.util.Map of the extra infos of this results.
   * 
   * @return the <b>internal</b> Map used by this results, modify with care.
   * @throws IllegalStateException if this method is invoked on instance of DBData
   * @since jcms-5.7.1
   */
  public HashMap<String, String> getExtraDataMap() {
    if (isDBData()) {
      throw new IllegalStateException("ExtraData cannot be used for DBData, use ExtraDBData instead.");
    }
    return extraDataMap;
  }

  /**
   * Override the current extraDataMap with the given one
   * 
   * @param extraDataMap the new extraDataMap
   * @throws IllegalStateException if this method is invoked on instance of DBData
   * @since jcms-5.7.1
   */
  public void setExtraDataMap(HashMap<String, String> extraDataMap) {
    if (isDBData()) {
      throw new IllegalStateException("ExtraData cannot be used for DBData, use ExtraDBData instead.");
    }
    this.extraDataMap = extraDataMap;
  }

  /**
   * Returns the value of the named extra data as an <code>String</code>,
   * or <code>null</code> if no extra info of the given name exists.
   * 
   * @param name a <code>String</code> specifying the name of the extra data to retrieve
   * @return previous value associated with specified name, or <tt>null</tt> if there is no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @throws IllegalStateException if this method is invoked on instance of DBData
   * @since jcms-5.7.1
   */
  public String getExtraData(String name) {
    if (isDBData()) {
      throw new IllegalStateException("ExtraData cannot be used for DBData, use ExtraDBData instead.");
    }
    if (extraDataMap == null) {
      return null;
    }

    return extraDataMap.get(name);
  }

  /**
   * Stores an extra data in this <code>Data</code>. <br>
   * If the object passed in is null, the effect is the same as
   * calling {@link #removeExtraData}.<br>
   * 
   * @param name a <code>String</code> specifying the name of the extra data
   * @param value the <code>String</code> to be stored
   * @return previous value associated with specified name, or <tt>null</tt> if there was no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @throws IllegalStateException if this method is invoked on instance of DBData
   * @since jcms-5.7.1
   */
  public String setExtraData(String name, String value) {
    if (isDBData()) {
      throw new IllegalStateException("ExtraData cannot be used for DBData, use ExtraDBData instead.");
    }
    if (value == null) {
      return removeExtraData(name);
    }

    if (extraDataMap == null) {
      extraDataMap = new HashMap<String, String>();
    }

    return extraDataMap.put(name, value);

  }

  /**
   * Removes an extra data from this <code>Data</code>.
   * 
   * @param name a <code>String</code> specifying the name of the extra data
   * @return previous value associated with specified name, or <tt>null</tt> if there was no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @throws IllegalStateException if this method is invoked on instance of DBData
   * @since jcms-5.7.1
   */
  public String removeExtraData(String name) {
    if (isDBData()) {
      throw new IllegalStateException("ExtraData cannot be used for DBData, use ExtraDBData instead.");
    }
    if (extraDataMap == null) {
      return null;
    }

    return extraDataMap.remove(name);
  }

  /**
   * Returns a <code>TreeMap</code> representing the Extra
   * Data description and default value.
   * Note: this implementation always create a new TreeMap because keys are sorted
   * 
   * @param clazz the class to work with
   * @param shortKey if true, returns a Map using short properties (someprop instead of someprop.boolean)
   * @return TreeMap object
   * @since jcms-5.7.1
   * @deprecated use {@link #getNewDefaultExtraDataMap(Class,String,boolean)}
   */
  @Deprecated
  public static TreeMap<String, String> getNewDefaultExtraDataMap(Class<? extends Data> clazz, boolean shortKey) {
    return getNewDefaultExtraDataMap(clazz, null, shortKey);
  }

  /**
   * Returns a <code>TreeMap</code> representing the Extra
   * Data description and default value.
   * Note: this implementation always create a new TreeMap because keys are sorted
   * 
   * @param clazz the class to work with
   * @param suffix a suffix to the properties
   * @param shortKey if true, returns a Map using short properties (someprop instead of someprop.boolean)
   * @return TreeMap object
   * @since jcms-5.7.4
   */
  public static TreeMap<String, String> getNewDefaultExtraDataMap(Class<? extends Data> clazz, String suffix, boolean shortKey) {

    if (clazz == null) {
      return null;
    }
    TreeMap<String, String> extraDataMap = new TreeMap<String, String>();

    String className = Util.getClassShortName(clazz);
    String prefix = "extra." + className + "." + (Util.isEmpty(suffix) ? "" : suffix); // bug JCMS-1627
    LangProperties props = Channel.getChannel().getProperties(prefix);

    // Directly return the Map
    if (!shortKey) {
      extraDataMap.putAll(props);
      return extraDataMap;
    }

    // Iterate on the map to return shortKey Map
    for (Map.Entry<String, String> entry : props.entrySet()) {
      String itKey = entry.getKey();
      String itDefault = entry.getValue();
      String itEditor = WidgetUtil.getPropertyEditor(itKey);
      String itShortKey = WidgetUtil.getShortProperty(itKey, itEditor);
      extraDataMap.put(itShortKey, itDefault);
    }

    return extraDataMap;
  }

  /**
   * Returns a <code>TreeMap</code> representing the Extra
   * Data description and default value.
   * Note: this implementation always create a new TreeMap because keys are sorted
   * 
   * @param clazz the class to work with
   * @param suffix a suffix to the properties
   * @param shortKey return a Map using short properties
   * @param useSuperClasses true to retrieve default extradata for super classes of the specified class
   * @return TreeMap object
   * @since jcms-5.7.5
   */
  @SuppressWarnings("unchecked")
  public static TreeMap<String, String> getNewDefaultExtraDataMap(Class<? extends Data> clazz, String suffix, boolean shortKey, boolean useSuperClasses) {
    if (!useSuperClasses) {
      return getNewDefaultExtraDataMap(clazz, suffix, shortKey);
    }

    if (clazz == null) {
      return null;
    }

    // Initialize with exact type
    TreeMap<String, String> extraDataMap = getNewDefaultExtraDataMap(clazz, suffix, shortKey);

    // Search in type hierarchy
    for (Class<? extends Data> curClazz = (Class<? extends Data>) clazz.getSuperclass(); curClazz != null; curClazz = (Class<? extends Data>) curClazz.getSuperclass()) {
      extraDataMap.putAll(getNewDefaultExtraDataMap(curClazz, suffix, shortKey));
      if (curClazz == Data.class) {
        break;
      }
    }

    return extraDataMap;
  }

  /**
   * Returns a <code>LangProperties</code> representing the Extra
   * Data description and default values.
   * 
   * @param shortKey if true, returns a Map using short properties (someprop instead of someprop.boolean)
   * @return LangProperties object
   * @throws IllegalStateException if this method is invoked on instance of DBData
   * @since jcms-5.7.1
   */
  public TreeMap<String, String> getNewDefaultExtraDataMap(boolean shortKey) {
    if (isDBData()) {
      throw new IllegalStateException("ExtraData cannot be used for DBData, use ExtraDBData instead.");
    }
    return getNewDefaultExtraDataMap(getClass(), null, shortKey, true);
  }

  /**
   * Update editable ExtraDBData of this data with the updated/new specified key/value <p>.
   * <p>
   * This method will NOT allow add/modification of non-editable extradbdata, ie : 
   * extradata which are not explicitely declared in properties.
   * <p>
   * This method is meant to be invoked on a working data (new object before 
   * its creation or updateinstance before its update).<p>
   * It is used Edit Handlers of all Data. 
   * <p>
   * Implementation detail : this method creates a new Map to simplify update
   * of a cloned JStore object (fix issue when extraDataMap has not been cloned
   * prior update)
   * 
   * @param keys String[] representing authorized properties
   * @param values String[] representing new values of properties
   * @throws IllegalStateException if this method is invoked on instance of DBData
   */
  public void updateExtraDataMap(String[] keys, String[] values) {
    if (isDBData()) {
      throw new IllegalStateException("ExtraData cannot be used for DBData, use ExtraDBData instead.");
    }

    if ((keys == null) || (values == null) || (keys.length != values.length)) {
      return;
    }
    
    // 1. Create new Map
    HashMap<String, String> newExtraDataMap = new HashMap<String, String>();

    // 2. Retrieve default key/value for editable extra data 
    TreeMap<String, String> defaultProps = getNewDefaultExtraDataMap(true);
    newExtraDataMap.putAll(defaultProps);
    
    // 3. Append previous unmodified extra data
    if (extraDataMap != null) {
      newExtraDataMap.putAll(extraDataMap);
    }

    // 4. Update/Overwrite with new given values
    for (int i = 0; i < keys.length; i++) {
      if (Util.isEmpty(keys[i])) {
        continue;
      }
      // Do not allow add/update of non editable extradata (ie: extradata not declared in properties)
      // This will prevent user to add any unwanted data 
      if (!defaultProps.containsKey(keys[i])) {
        continue;
      }     
      newExtraDataMap.put(keys[i], Util.getString(values[i], ""));
    }
    
    // Reset the extraDataMap
    this.extraDataMap = newExtraDataMap;
  }

  // ----------------------------------------------------------------------
  // Extra DB Data
  // ----------------------------------------------------------------------

  transient protected Map<String, String> extraDBDataMapToSaveOnOp;

  /**
   * Checks the storedstate of this data to determine which extradbdata storage system should be used (immediate vs internal cache).
   * <p>
   * Returns :
   * <ul>
   * <li><code>true</code>, when this Data is in a StoredState allowing an immediate access to DB.</li>
   * <li><code>false</code>, when this Data is a working copy indicating that the internal extraDBData cache should be used.</li>
   * <ul>
   * 
   * @return true if immediate access to DB should be used false to used object's extradbdata cache.
   * @since jcms-6.0
   */
  private boolean extraDBDataImmediateAccess() {
    StoredState ss = getStoredState();
    switch (ss) {
    // Save for next operation
    case DATA_CLONED:
    case DATA_NEW:
    case DBDATA_NEW: {
      return false;
    }

      // Save immediately
    case DATA_ORIGINAL:
    case DBDATA_STORED:
    default: {
      return true;
    }
    }
  }

  /**
   * Returns a java.util.Map of the extra infos of this results.
   * 
   * @return the <b>internal and unmodifiable</b> Map.
   * @since jcms-6.0
   */
  public Map<String, String> getExtraDBDataMap() {
    Map<String, String> returnedMap = null;

    if (extraDBDataMapToSaveOnOp == null && extraDBDataImmediateAccess()) {
      returnedMap = ExtraDBDataManager.getExtraDBDataMap(this);
    } else {
      returnedMap = extraDBDataMapToSaveOnOp;
    }
    return returnedMap == null ? null : Collections.unmodifiableMap(returnedMap);
  }

  /**
   * Override the current extraDBDataMap with the given one
   * 
   * @param extraDBDataMap the new extraDBDataMap
   * @since jcms-6.0
   */
  public void setExtraDBDataMap(Map<String, String> extraDBDataMap) {
    if (extraDBDataImmediateAccess()) {
      ExtraDBDataManager.saveExtraDBDataMap(this, extraDBDataMap, true);
    } else {
      extraDBDataMapToSaveOnOp = extraDBDataMap;
    }
  }
  
  /**
   * Override the current extraDBDataMap with the given one
   * 
   * @param extraDBDataMap the new extraDBDataMap
   * @param inMemoryOnly if specified to true, the map will be kept in memory (as a ariable of this Data)
   *        variable <b>and will only be saved if invoking performupdate later</b> 
   * @since jcms-7.0.1
   */
  public void setExtraDBDataMap(Map<String, String> extraDBDataMap, boolean inMemoryOnly) {
    if (!inMemoryOnly && extraDBDataImmediateAccess()) {
      ExtraDBDataManager.saveExtraDBDataMap(this, extraDBDataMap, true);
    } else {
      extraDBDataMapToSaveOnOp = extraDBDataMap;
    }
  }

  /**
   * Returns the value of the named extra data as an <code>String</code>,
   * or <code>null</code> if no extra info of the given name exists.
   * 
   * @param name a <code>String</code> specifying the name of the extra data to retrieve
   * @return previous value associated with specified name, or <tt>null</tt> if there is no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @since jcms-6.0
   */
  public String getExtraDBData(String name) {
    if (extraDBDataMapToSaveOnOp == null && extraDBDataImmediateAccess()) {
      return ExtraDBDataManager.getExtraDBDataValue(this, name);
    }
    if (extraDBDataMapToSaveOnOp == null) {
      return null;
    }
    return extraDBDataMapToSaveOnOp.get(name);
  }

  /**
   * Stores an extra data in this <code>Data</code>. <br>
   * If the object passed in is null, the effect is the same as
   * calling {@link #removeExtraDBData(String)}.<br>
   * 
   * @param name a <code>String</code> specifying the name of the extra data
   * @param value the <code>String</code> to be stored
   * @return previous value associated with specified name, or <tt>null</tt> if there was no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @since jcms-6.0
   */
  public String setExtraDBData(String name, String value) {
    if (extraDBDataMapToSaveOnOp == null && extraDBDataImmediateAccess()) {
      return ExtraDBDataManager.saveExtraDBData(this, name, value);
    }
    if (extraDBDataMapToSaveOnOp == null) {
      extraDBDataMapToSaveOnOp = new HashMap<String, String>();
    }
    return extraDBDataMapToSaveOnOp.put(name, value);
  }

  /**
   * Removes an extra data from this <code>Data</code>.
   * 
   * @param name a <code>String</code> specifying the name of the extra data
   * @return previous value associated with specified name, or <tt>null</tt> if there was no mapping for name. A <tt>null</tt> return can
   * also indicate that <tt>null</tt> was associated
   * with the specified name.
   * @since jcms-6.0
   */
  public String removeExtraDBData(String name) {
    return setExtraDBData(name, null);
  }

  /**
   * Returns a <code>TreeMap</code> representing the Extra
   * Data description and default value.
   * Note: this implementation always create a new TreeMap because keys are sorted
   * 
   * @param clazz the class to work with
   * @param shortKey if true, returns a Map using short properties (someprop instead of someprop.boolean)
   * @return TreeMap object
   * @since jcms-6.0
   * @deprecated use {@link #getNewDefaultExtraDBDataMap(Class,String,boolean)}
   */
  @Deprecated
  public static TreeMap<String, String> getNewDefaultExtraDBDataMap(Class<? extends Data> clazz, boolean shortKey) {
    return getNewDefaultExtraDBDataMap(clazz, null, shortKey);
  }

  /**
   * Returns a <code>TreeMap</code> representing the Extra
   * Data description and default value.
   * Note: this implementation always create a new TreeMap because keys are sorted
   * 
   * @param clazz the class to work with
   * @param suffix a suffix to the properties
   * @param shortKey if true, returns a Map using short properties (someprop instead of someprop.boolean)
   * @return TreeMap object
   * @since jcms-6.0
   */
  public static TreeMap<String, String> getNewDefaultExtraDBDataMap(Class<? extends Data> clazz, String suffix, boolean shortKey) {

    if (clazz == null) {
      return null;
    }
    TreeMap<String, String> extraDBDataMap = new TreeMap<String, String>();

    String className = Util.getClassShortName(clazz);
    String prefix = "extradb." + className + "." + (Util.isEmpty(suffix) ? "" : suffix); // bug JCMS-1627
    LangProperties props = Channel.getChannel().getProperties(prefix);

    // Directly return the Map
    if (!shortKey) {
      extraDBDataMap.putAll(props);
      return extraDBDataMap;
    }

    // Iterate on the map to return shortKey Map
    for (Map.Entry<String, String> entry : props.entrySet()) {
      String itKey = entry.getKey();
      String itDefault = entry.getValue();
      String itEditor = WidgetUtil.getPropertyEditor(itKey);
      String itShortKey = WidgetUtil.getShortProperty(itKey, itEditor);
      extraDBDataMap.put(itShortKey, itDefault);
    }

    return extraDBDataMap;
  }

  /**
   * Returns a <code>TreeMap</code> representing the Extra DB
   * Data description and default value.
   * Note: this implementation always create a new TreeMap because keys are sorted
   * 
   * @param clazz the class to work with
   * @param suffix a suffix to the properties
   * @param shortKey return a Map using short properties
   * @param useSuperClasses true to retrieve default extradbdata for super classes of the specified class
   * @return TreeMap object
   * @since jcms-6.0
   */
  @SuppressWarnings("unchecked")
  public static TreeMap<String, String> getNewDefaultExtraDBDataMap(Class<? extends Data> clazz, String suffix, boolean shortKey, boolean useSuperClasses) {
    if (!useSuperClasses) {
      return getNewDefaultExtraDBDataMap(clazz, suffix, shortKey);
    }

    if (clazz == null) {
      return null;
    }

    // Initialize with exact type
    TreeMap<String, String> extraDBDataMap = getNewDefaultExtraDBDataMap(clazz, suffix, shortKey);

    // Search in type hierarchy
    for (Class<? extends Data> curClazz = (Class<? extends Data>) clazz.getSuperclass(); curClazz != null; curClazz = (Class<? extends Data>) curClazz.getSuperclass()) {
      extraDBDataMap.putAll(getNewDefaultExtraDBDataMap(curClazz, suffix, shortKey));
      if (curClazz == Data.class) {
        break;
      }
    }

    return extraDBDataMap;
  }

  /**
   * Returns a <code>LangProperties</code> representing the Extra
   * Data description and default values.
   * 
   * @param shortKey if true, returns a Map using short properties (someprop instead of someprop.boolean)
   * @return LangProperties object
   * @since jcms-6.0
   */
  public TreeMap<String, String> getNewDefaultExtraDBDataMap(boolean shortKey) {
    return getNewDefaultExtraDBDataMap(getClass(), null, shortKey, true);
  }

  /**
   * Update editable ExtraDBData of this data with the updated/new specified key/value <p>.
   * <p>
   * This method will NOT allow add/modification of non-editable extradbdata, ie : 
   * extradata which are not explicitely declared in properties.
   * <p>
   * This method is meant to be invoked on a working data (new object before 
   * its creation or updateinstance before its update).<p>
   * It is used Edit Handlers of all Data. 
   * 
   * @param keys String[] representing authorized properties
   * @param values String[] representing new values of properties
   * @since jcms-6.0
   */
  public void updateExtraDBDataMap(String[] keys, String[] values) {

    if ((keys == null) || (values == null) || (keys.length != values.length)) {
      return;
    }
    
    // 1. Create new Map
    HashMap<String, String> newExtraDBDataMap = new HashMap<String, String>();

    // 2. Retrieve default key/value for editable extra db data 
    TreeMap<String, String> defaultProps = getNewDefaultExtraDBDataMap(true);
    newExtraDBDataMap.putAll(defaultProps);
    
    // 3. Append previous unmodified extra data
    Map<String, String> extraDBDataMap = getExtraDBDataMap();
    if (extraDBDataMap != null) {
      newExtraDBDataMap.putAll(extraDBDataMap);
    }

    // 4. Update/Overwrite with new given values
    for (int i = 0; i < keys.length; i++) {
      if (Util.isEmpty(keys[i])) {
        continue;
      }
      // Do not allow add/update of non editable extradata (ie: extradata not declared in properties)
      // This will prevent user to add any unwanted data 
      if (!defaultProps.containsKey(keys[i])) {
        continue;
      }    
      newExtraDBDataMap.put(keys[i], Util.getString(values[i], ""));
    }

    // Reset the extraDBDataMap
    setExtraDBDataMap(newExtraDBDataMap, true);
  }

  // ----------------------------------------------------------------------
  // Lock management
  // ----------------------------------------------------------------------

  /**
   * Locks this data for a given member.
   * 
   * @param mbr the member who locks this data
   * @since jcms-5.5.0
   */
  public void lock(Member mbr) {
    WeakLockManager.getInstance().lock(this, mbr);
  }

  /**
   * Unlocks this data if lock has been put by the given member.
   * 
   * @param mbr the member who locked this data
   * @since jcms-5.5.0
   */
  public void unlock(Member mbr) {
    WeakLockManager.getInstance().unlock(this, mbr);
  }

  /**
   * Unlocks this data.
   * 
   * @since jcms-5.5.0
   */
  public void unlock() {
    WeakLockManager.getInstance().unlock(this);
  }

  /**
   * Return the member who locked this data
   * 
   * @return the member who locked this data
   * @since jcms-5.5.0
   */
  public Member getLockMember() {
    return WeakLockManager.getInstance().getLockMember(this);
  }

  /**
   * Returns the lock date.
   * 
   * @return the lock date
   * @since jcms-5.5.0
   */
  public Date getLockDate() {
    return WeakLockManager.getInstance().getLockDate(this);
  }

  /**
   * Returns true if this data is locked.
   * 
   * @return true if this data is locked.
   * @since jcms-5.5.0
   */  public boolean isLocked() {
    return WeakLockManager.getInstance().isLocked(this);
  }

  /**
   * Returns true if the given member has locked the given data or if the data is not locked.
   * 
   * @param mbr the member
   * @return true if the given member has locked the given data or if the data is not locked.
   * @since jcms-7.0.0
   */
  public boolean checkLockMember(Member mbr) {
    return WeakLockManager.getInstance().checkLockMember(this, mbr);
  }

  
  // ----------------------------------------------------------------------
  // Strong Lock management
  // ----------------------------------------------------------------------
  /**
   * Returns the member who has strongly locked this Data.
   * @return the member who has strongly locked this Data.
   * @since jcms-7.0.0
   */
  public Member getStrongLockMember() {
    StrongLock lock = getStrongLock();
    if (lock == null) {
      return null;
    }
    return lock.getAuthor();
  }
  
  /**
   * Returns the date of the strong lock.
   * @return the date of the strong lock.
   * @since jcms-7.0.0
   */
  public Date getStrongLockDate() {
    StrongLock lock = getStrongLock();
    if (lock == null) {
      return null;
    }
    return lock.getCdate();
  }
  
  /**
   * Put (and save) a strong lock on this Data.
   * 
   * @param mbr the member who puts the strong lock.
   * @since jcms-7.0.0
   */
  public void putStrongLock(Member mbr) {
    StrongLockManager.getInstance().lock(this, mbr);
  }
 

  /**
   * Release the lock on this Data
   * 
   * @param mbr the member who releases the lock 
   * (only the member who has locked and the admin 
   * can release the lock).
   * @since jcms-7.0.0
   */
  public void releaseStrongLock(Member mbr) {
    StrongLockManager.getInstance().unlock(this, mbr);
  }
  

  /**
   * Releases the strong lock, without checking the member.
   * 
   * @since jcms-7.0.0
   */
  public void releaseStrongLock() {
    StrongLockManager.getInstance().unlock(this);
  }
  
  /**
   * Returns a string which explains who has the lock (return "" if this Data has no strong lock)
   * @param locale the locale
   * @return a string which explains who has the lock.
   * @since jcms-7.0.0
   */
  public String getStrongLockInfo(Locale locale) {
    return StrongLockManager.getInstance().getLockInfo(this, locale);
  }
  
  
  /**
   * Returns true if the given member can request to unlock this data.
   * @param mbr the member who request the lock
   * @return true if the given member can request to unlock this data.
   * @since jcms-8.0.0
   */
  public boolean canRequestUnlock(Member mbr) {
    return StrongLockManager.getInstance().canRequestUnlock(this, mbr);
  }
  
  /**
   * Returns true if the given member can unlock this data.
   * @param mbr the member
   * @return true if the given member can unlock this data.
   * @since jcms-9.0.0
   */
  public boolean canBeUnlockedBy(Member mbr) {
    return StrongLockManager.getInstance().canUnlock(this, mbr);
  }  
  
  /**
   * Returns the StrongLock bound to this Data, if any.
   * 
   * @return the StrongLock instance if there is any for this Data, null otherwise
   * @see StrongLockManager#getStrongLock(Data)
   * @since jcms-6.2.0
   */
  public StrongLock getStrongLock() {
    return StrongLockManager.getInstance().getStrongLock(this);
  }

  /**
   * Check if this data is strongly locked.
   * 
   * @return true if this Data has a StrongLock, false otherwise
   * @see StrongLockManager#isLocked(Data) 
   * @since jcms-6.2.0
   */
  public boolean isStrongLocked() {
    return StrongLockManager.getInstance().isLocked(this);
  }

  /**
   * Check if the specified member has locked this Data or if the Data is not locked.
   * 
   * @param mbr the member to check
   * @return true if the given member has locked the given data or if the data is not locked.
   * @see StrongLockManager#checkLockMember(Data, Member)
   * @since jcms-6.2.0
   */
  public boolean checkStrongLockMember(Member mbr) {
    return StrongLockManager.getInstance().checkLockMember(this, mbr);
  }

  /**
   * Returns an information message about the lock state.
   * 
   * @param userLang the user language
   * @param userLocale the user locale
   * @param confirm true for a confirm message
   * @return an information message about the lock state.
   * @since jcms-6.2.0
   */
  public String getLockInfo(String userLang, Locale userLocale, boolean confirm) {
    // Get lock info
    Member owner = getLockMember();
    Date date = getLockDate();
    StrongLock strongLock = getStrongLock();
    if (strongLock != null) {
      owner = strongLock.getAuthor();
      date = strongLock.getCdate();
    }

    // Compute lock message
    String ownerLink = "<a href=" + JcmsUtil.getDisplayUrl(owner, userLocale) + ">" + owner.getFullName() + "</a>";
    String dateStr = DateFormat.getDateInstance(DateFormat.SHORT, userLocale).format(date);
    String timeStr = DateFormat.getTimeInstance(DateFormat.SHORT, userLocale).format(date);
    String info = null;
    if (strongLock != null) {
      String prop = confirm ? "ui.work.form.txt.lock-strong.confirm" : "ui.work.form.txt.lock-strong";
      info = JcmsUtil.glp(userLang, prop, ownerLink, dateStr, timeStr);
    } else {
      info = JcmsUtil.glp(userLang, "ui.work.form.txt.lock", ownerLink, timeStr);
    }

    return info;
  }

  // ----------------------------------------------------------------------
  // DataController
  // ----------------------------------------------------------------------

  /**
   * Returns an instance of this object to update it.
   * 
   * @return an instance of this object to update it.
   * @since jcms-6.0.0
   */
  public Data getUpdateInstance() {
    if (isDBData()) {
      return this;
    }
    return (Data) this.clone();
  }
  
  // ReFix bug JCMS-2655 : checkIntegrity method is invoked twice per Data
  // 
  // This ThreadLocale is used to check when checkIntegrity(Map) must be invoked from checkIntegrity()
  //
  // Expected Behavior :
  //
  //   checkIntegrity()
  //     -> checkIntegrity(map{legacy=true}) Unless ThreaLocal==false
  //       -> DataController.checkIntegrity
  //
  //   checkIntegrity(map{legacy=null==false})
  //     -> checkIntegrity()  ThreadLocal=false
  //     -> DataController.checkIntegrity
  //
  private static ThreadLocal<Boolean> invokeCheckIntegrityWithMapTL = new ThreadLocal<Boolean>();

  /**
   * Checks whether or not data integrity constraints are respected.
   * 
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkIntegrity() {
    
    // Do not invoke checkIntegrity(Map) method when checkIntegrity() was invoked from checkIntegrity(Map)
    boolean invokeCheckIntegrityWithMap = Util.toBoolean(invokeCheckIntegrityWithMapTL.get(), true);
    if (invokeCheckIntegrityWithMap) {
      Map<Object, Object> context = new HashMap<Object, Object>();
      context.put("isLegacyCheckIntegrity", Boolean.TRUE);
      return checkIntegrity(context);  
    }
    
    return ControllerStatus.OK;
  }
  /**
   * Checks whether or not data integrity constraints are respected.
   * 
   * @param context a map which contains context parameters (may be null)
   * @return a ControllerStatus
   * @since jcms-6.1.0
   */
  public ControllerStatus checkIntegrity(Map context) {
    // Ensure that legacy method checkIntegrity() is invoked
    boolean isDirectInvokationWithoutMap = context == null;
    boolean isLegacyInvokation = context != null && Util.toBoolean(context.get("isLegacyCheckIntegrity"), false);
    if (isDirectInvokationWithoutMap || !isLegacyInvokation) {
      ControllerStatus status = ControllerStatus.OK;
      try {
        invokeCheckIntegrityWithMapTL.set(Boolean.FALSE);
        status = checkIntegrity();
      } finally {
        invokeCheckIntegrityWithMapTL.set(Boolean.TRUE);
      }
      if (status.hasFailed()) {
        return status;
      }
    }
    
    // Check custom controlers
    List classList = Util.getSuperclassList(getClass(), Data.class, true, true);
    for (Iterator it = classList.iterator(); it.hasNext();) {
      Class itClass = (Class) it.next();
      Set dcSet = channel.getDataControllerSet(itClass);
      if (Util.isEmpty(dcSet)) {
        continue;
      }
      for (Iterator it2 = dcSet.iterator(); it2.hasNext();) {
        DataController dc = (DataController) it2.next();
        ControllerStatus dcStatus = dc.checkIntegrity(this);
        if (dcStatus == null) {
          logger.warn("DataController.checkIntegrity '" + Util.getClassShortName(dc.getClass()) + "' may return null Controller status");
        } else if (dcStatus.hasFailed()) {
          return dcStatus;
        }
      }
    }
      
    return ControllerStatus.OK;
  }

  /**
   * Returns the FieldStatus map of this data for the given member (currently only works for Publication).
   * 
   * @param mbr the member
   * @return the FieldStatus map (the field names are the keys)
   * @since jcms-6.0.1
   */
  public Map<String, FieldStatus> getFieldStatusMap(Member mbr) {
    HashMap<String, FieldStatus> map = new HashMap<String, FieldStatus>();
    // Check custom controlers
    List classList = Util.getSuperclassList(getClass(), Data.class, true, true);
    for (Iterator it = classList.iterator(); it.hasNext();) {
      Class itClass = (Class) it.next();
      Set dcSet = channel.getDataControllerSet(itClass);
      if (Util.isEmpty(dcSet)) {
        continue;
      }
      for (Iterator it2 = dcSet.iterator(); it2.hasNext();) {
        DataController dc = (DataController) it2.next();
        dc.processFieldStatusMap(this, mbr, map);
      }
    }
    return map;
  }

  /**
   * Checks if the given write operation can be performed
   * 
   * @param op the operation (OP_CREATE, OP_UPDATE, ...)
   * @param mbr the member which requests the write operation
   * @param checkIntegrity if true, requires the constraints integrity to be checked
   * @param context a map which contains context parameters (may be null)
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkWrite(int op, Member mbr, boolean checkIntegrity, Map context) {
    ControllerStatus status = null;
    
    // Check data integrity
    if (checkIntegrity && (context == null || !context.containsKey(DataController.CTXT_IGNORE_INTEGRITY_CHECK))) {
      status = checkIntegrity(context);
      if (status.hasFailed()) {
        return status;
      }
    }

    // Check if data write enabled
    if (!channel.isDataWriteEnabled()) {
      status = new ControllerStatus(CTRL_TOPIC_WRITE);
      status.setProp("channel.data-write.status");
      return status;
    }

    // Check member rights
    status = checkMember(op, mbr, context);
    if (status.hasFailed()) {
      return status;
    }

    // Check custom controlers
    List classList = Util.getSuperclassList(getClass(), Data.class, true, true);
    for (Iterator it = classList.iterator(); it.hasNext();) {
      Class itClass = (Class) it.next();
      Set dcSet = channel.getDataControllerSet(itClass);
      if (Util.isEmpty(dcSet)) {
        continue;
      }
      for (Iterator it2 = dcSet.iterator(); it2.hasNext();) {
        DataController dc = (DataController) it2.next();
        status = dc.checkWrite(this, op, mbr, checkIntegrity, context);
        if (status == null) {
          logger.warn("DataController.checkWrite '" + Util.getClassShortName(dc.getClass()) + "' may return null Controller status");
        } else if (status.hasFailed()) {
          return status;
        }
      }
    }

    return ControllerStatus.OK;
  }

  private static final Map<?, ?> getDefaultContext() {
    return new HashMap();
  }
  
  /**
   * Checks if the given member can perform the given operation. This
   * method is called by checkWrite. It should be overrided by
   * sub-classes.
   * 
   * @param op the operation (OP_CREATE, OP_UPDATE, ...)
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkMember(int op, Member mbr, Map context) {
    return ControllerStatus.OK;
  }

  /**
   * Checks if the creation can be performed
   * 
   * @param mbr the member which requests the write operation
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkCreate(Member mbr) {
    return checkCreate(mbr, getDefaultContext());
  }

  /**
   * Checks if the creation can be performed
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkCreate(Member mbr, Map context) {
    return checkWrite(OP_CREATE, mbr, true, context);
  }

  /**
   * Checks if the update can be performed
   * 
   * @param mbr the member which requests the write operation
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkUpdate(Member mbr) {
    return checkUpdate(mbr, getDefaultContext());
  }

  /**
   * Checks if the update can be performed
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkUpdate(Member mbr, Map context) {
    return checkWrite(OP_UPDATE, mbr, true, context);
  }

  /**
   * Checks if the deletion can be performed
   * 
   * @param mbr the member which requests the write operation
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkDelete(Member mbr) {
    return checkDelete(mbr, getDefaultContext());
  }

  /**
   * Checks if the deletion can be performed
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @return a ControllerStatus
   * @since jcms-5.0.0
   */
  public ControllerStatus checkDelete(Member mbr, Map context) {
    Set<String> vidSet = getVirtualIDSet();
    if (Util.notEmpty(vidSet) && Boolean.TRUE != context.get(DataController.CTXT_FORCE_DELETE)) {
      ControllerStatus status = new ControllerStatus(CTRL_TOPIC_INTERNAL);
      status.setProp("msg.edit.delete.vid-exist", vidSet.size(), Util.join(vidSet, ", "));
      return status;
    }
    return checkWrite(OP_DELETE, mbr, false, context);
  }

  /**
   * Returns the set of virtual ID bound to this data (or null if it does not exist).
   * 
   * @return the set of virtual ID bound to this data (or null if it does not exist).
   * @since jcms-8.0.0
   */
  public Set<String> getVirtualIDSet() {
    return channel.getVirtualIDSet(this);
  }

  /**
   * Checks if the creation can be performed, and performs data creation. This method cannot be overridden.
   * 
   * @param mbr the member which requests the write operation
   * @return a {@link ControllerStatus}
   */
  public final ControllerStatus checkAndPerformCreate(Member mbr) {
    return checkAndPerformCreate(mbr, getDefaultContext());
  }

  /**
   * Checks if the creation can be performed, and performs data creation.
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @return a {@link ControllerStatus}
   */
  public ControllerStatus checkAndPerformCreate(Member mbr, Map<?, ?> context) {
    ControllerStatus status = checkCreate(mbr, context);
    if (status.isOK()) {
      performCreate(mbr, context);
    }
    return status;
  }
  
  /**
   * Performs data creation. This method cannot be overridden.
   * 
   * @param mbr the member which requests the write operation
   * @since jcms-5.0.0
   */
  public final void performCreate(Member mbr) {
    performCreate(mbr, getDefaultContext());
  }

  /**
   * Performs data creation.
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @since jcms-5.0.0
   */
  public void performCreate(Member mbr, Map context) {
    if (controlBeforeWrite(OP_CREATE, mbr, context)) {
      channel.createData(this, mbr);
      if (extraDBDataMapToSaveOnOp != null) {
        ExtraDBDataManager.saveExtraDBDataMap(this, extraDBDataMapToSaveOnOp, true);
        extraDBDataMapToSaveOnOp = null;
      }
    }
    controlAfterWrite(OP_CREATE, mbr, context);
  }

  /**
   * Checks if the update can be performed, and performs data update. This method cannot be overridden.
   * 
   * @param mbr the member which requests the write operation
   * @return a {@link ControllerStatus}
   */
  public final ControllerStatus checkAndPerformUpdate(Member mbr) {
    return checkAndPerformUpdate(mbr, getDefaultContext());
  }
  
  /**
   * Checks if the update can be performed, and performs data update.
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @return a {@link ControllerStatus}
   */
  public ControllerStatus checkAndPerformUpdate(Member mbr, Map<?, ?> context) {
    ControllerStatus status = checkUpdate(mbr, context);
    if (status.isOK()) {
      performUpdate(mbr, context);
    }
    return status;
  }
  
  /**
   * Performs data update. This method cannot be overridden.
   * 
   * @param mbr the member which requests the write operation
   * @since jcms-5.0.0
   */
  public final void performUpdate(Member mbr) {
    performUpdate(mbr, getDefaultContext());
  }

  /**
   * Performs data update.
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @since jcms-5.0.0
   */
  public void performUpdate(Member mbr, Map context) {
    
    if (isDBData()) {
      initializeDBDataCollections();
    }
    
    prepareUpdateContext(context);

    if (controlBeforeWrite(OP_UPDATE, mbr, context)) {
      channel.updateData(this, mbr);
      if (extraDBDataMapToSaveOnOp != null) {
        ExtraDBDataManager.saveExtraDBDataMap(this, extraDBDataMapToSaveOnOp, false);
        extraDBDataMapToSaveOnOp = null;
      }
    }
    controlAfterWrite(OP_UPDATE, mbr, context);
    
    // Fix bug JCMS-3397: Descriptive URL of DBData may still temporarly display the previous url after a data update
    if (isDBData()) {
      DescriptiveURLs.removeURLCache(this);
    }
  }

  public void prepareUpdateContext(Map context) {
    if (context == null) {
      return;
    }

    Data previousData = (Data) context.get(DataController.CTXT_PREVIOUS_DATA);
    if (JcmsUtil.isSameId(previousData, this)) {
      return;
    }

    // Add previous data
    try {
      if (isDBData()) {
        previousData = HibernateUtil.getCurrentData(getId());
      } else {
        previousData = (Data) channel.getData(id).clone();
      }
      context.put(DataController.CTXT_PREVIOUS_DATA, previousData);

      // Improvement JCMS-4815 : On DBData update, provide the previous DBData instance in extrainfo for use in DBListener
      if (isDBData()) {
        setExtraInfo(DataController.CTXT_PREVIOUS_DATA, previousData);
      }
      
    } catch (Throwable th) {
      // EMPTY
    }
  }

  /**
   * Checks if the deletion can be performed, and performs data deletion. This method cannot be overridden.
   * 
   * @param mbr the member which requests the write operation
   * @return a {@link ControllerStatus}
   */
  public final ControllerStatus checkAndPerformDelete(Member mbr) {
    return checkAndPerformDelete(mbr, getDefaultContext());
  }
  
  /**
   * Checks if the deletion can be performed, and performs data deletion.
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @return a {@link ControllerStatus}
   */
  public ControllerStatus checkAndPerformDelete(Member mbr, Map<?, ?> context) {
    ControllerStatus status = checkDelete(mbr, context);
    if (status.isOK()) {
      performDelete(mbr, context);
    }
    return status;
  }
  
  /**
   * Performs data deletion. This method cannot be overridden.
   * 
   * @param mbr the member which requests the write operation
   * @since jcms-5.0.0
   */
  public final void performDelete(Member mbr) {
    performDelete(mbr, getDefaultContext());
  }

  /**
   * Performs data deletion.
   * 
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @since jcms-5.0.0
   */
  public void performDelete(Member mbr, Map context) {

    // Delete the extension
    if (extension != null) {
      extension.performDelete(mbr, context);
    }
    
    if (isDBData()) {
      initializeDBDataCollections();
    }

    // Delete the object
    if (controlBeforeWrite(OP_DELETE, mbr, context)) {
      channel.deleteData(this, mbr);
      ExtraDBDataManager.removeAllExtraDBData(this);
      StrongLockManager.getInstance().unlock(this);
    }
    controlAfterWrite(OP_DELETE, mbr, context);
  }

  boolean controlBeforeWrite(int op, Member mbr, Map context) {
    controlWrite(op, mbr, context, true);
    performBeforeWrite(op, mbr, context);
    this.previousMDate = this.mdate == null ? 0L: this.mdate.getTime();
    return context == null || !Boolean.TRUE.equals(context.get(DataController.DO_NOT_STORE));    
  }

  void controlAfterWrite(int op, Member mbr, Map context) {
    
    controlWrite(op, mbr, context, false);
    performAfterWrite(op, mbr, context);
    
    //log event only if it's a creation, a delete or if document have modified mdate
    Data oldData = (Data) context.get(DataController.CTXT_PREVIOUS_DATA);
    boolean isDelete = op == OP_DEEP_DELETE || op == OP_DELETE;
    boolean isRealUpdate = false;
    if(op == OP_UPDATE || op == OP_MERGE){
      long newMDate = (channel.getData(this.getId()) == null  || channel.getData(this.getId()).mdate == null) ? 0L : channel.getData(this.getId()).mdate.getTime();
      isRealUpdate = this.previousMDate != newMDate;
    }
    if( op == OP_CREATE || op == OP_DEEP_COPY || isDelete || isRealUpdate){
      EventData eventData = new EventData(AnalyticsManager.CORE,mbr,op, this);
      if(this instanceof Member){
        Member mbrData = (Member) this;        
        completeContextWithMemberState(mbrData, oldData, isDelete, eventData);
      }
      try {
        AnalyticsManager.appendEventData(eventData, channel.getCurrentServletRequest());
      } catch (Exception e) {
        logger.warn(e.getMessage(), e);
      }
    }
  }

  private void completeContextWithMemberState( Member mbrData, Data oldData, boolean isDelete, EventData eventData) {
    Map<String, String> contextMap = eventData.getContext();
    String guestGroupId = channel.getProperty(AccountReportAgent.GUEST_ACCOUNTS_GROUP_ID_PROP);      
    Group guestGroup = Util.isEmpty(guestGroupId)? null : channel.getGroup(guestGroupId);      
    
    //add workspaces of the Member
    contextMap.put(AnalyticsManager.DATA_WORKSPACES, JcmsUtil.dataListToString(mbrData.getWorkspaceSet(), ","));
    
    if(mbrData.isAccount()){
      contextMap.put(AccountReportAgent.IS_ACCOUNT_KEY, "true");          
    }
    else if(mbrData.isContact()){
      contextMap.put(AccountReportAgent.IS_CONTACT_KEY, "true");
    }
    if( mbrData.isDisabled()){
      contextMap.put(AccountReportAgent.IS_DISABLED_KEY, "true");
    }
    if(guestGroup != null){
      if(mbrData.belongsToGroup(guestGroup)){
        contextMap.put(AccountReportAgent.IS_GUEST_KEY, "true");
      }
    }
    if(!isDelete && oldData != null && oldData instanceof Member){
      Member oldMbr = (Member) oldData; 
      contextMap.put(AnalyticsManager.OLD_DATA_WORKSPACES, JcmsUtil.dataListToString(oldMbr.getWorkspaceSet(), ","));
      if(oldMbr.isAccount()){
        contextMap.put(AccountReportAgent.IS_OLD_ACCOUNT_KEY, "true");          
      }
      else if(oldMbr.isContact()){
        contextMap.put(AccountReportAgent.IS_OLD_CONTACT_KEY, "true");
      }
      if( oldMbr.isDisabled()){
        contextMap.put(AccountReportAgent.IS_OLD_DISABLED_KEY, "true");
      }
      if(guestGroup != null){
        if(oldMbr.belongsToGroup(guestGroup)){
          contextMap.put(AccountReportAgent.IS_OLD_GUEST_KEY, "true");
        }
      }
    }
  }

  private void controlWrite(int op, Member mbr, Map context, boolean isBefore) {
    // Check custom controlers
    List<Class> classList = Util.getSuperclassList(getClass(), Data.class, true, true);
    if (Util.isEmpty(classList)) {
      return;
    }
    for (Class itClass: classList) {
      Set<DataController> dcSet = channel.getDataControllerSet(itClass);
      if (Util.isEmpty(dcSet)) {
        continue;
      }
      
      for (DataController dc: dcSet) {
        try {
          if (isBefore) {
            dc.beforeWrite(this, op, mbr, context);
          } else {
            dc.afterWrite(this, op, mbr, context);
          }
        }
        catch(Throwable th) {
          logger.warn("An exception occurred when calling method " + (isBefore ? "beforeWrite()" : "afterWrite()") + " for DataController " + dc, th);
        }
      }
    }
  }

  /**
   * Called to perform some action before the write.
   * 
   * @param op the operation (OP_CREATE, OP_UPDATE, ...)
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @since jcms-5.7.2
   */
  protected void performBeforeWrite(int op, Member mbr, Map context) {

    // Imported Data: mark as updated if this update is not a new import
    if (op != OP_UPDATE) {
      return;
    }

    if (!isImported()) {
      return;
    }

    if (isNewImport) {
      return;
    }

    if (context != null && IMPORT_STEP_UPDATE_REF == Util.toInt(context.get(IMPORT_DC_STEP), -1)) {
      return;
    }

    setImportUpdated(true);
  }

  /**
   * Called to perform some action after the write.
   * 
   * @param op the operation (OP_CREATE, OP_UPDATE, ...)
   * @param mbr the member which requests the write operation
   * @param context a map which contains context parameters (may be null)
   * @since jcms-6.0.0
   */
  protected void performAfterWrite(int op, Member mbr, Map context) {
    // EMPTY
  }

  // ----------------------------------------------------------------------
  // TYPE AND FIELD INFOS
  // ----------------------------------------------------------------------

  /**
   * Returns the TypeEntry bound to <code>Data</code>. <br>
   * <br>
   * Default implementation for <code>Data</code> will propably return null
   * as no TypeEntry is available for <code>Data</code>, this method
   * is available mainly to be overriden by subclasses. <br>
   * 
   * @return the TypeEntry bound to <code>Data</code>.
   * @see com.jalios.jcms.Channel#getTypeEntry(Class)
   * @since jcms-5.5
   */
  public static TypeEntry getTypeEntry() {
    return channel.getTypeEntry(Publication.class);
  }

  /**
   * Return an array of <code>TypeFieldEntry</code> bound to <code>Data</code>. <br>
   * This array <b>does not</b> contain inherited TypeFieldEntry from super classes of <code>Data</code>. <br>
   * <br>
   * Default implementation for <code>Data</code> will propably return an empty
   * array as no TypeFieldEntry is available for <code>Data</code>, this method
   * is available mainly to be overriden by subclasses. <br>
   * 
   * @return an array of <code>TypeFieldEntry</code> containing all Field used by <code>Data</code> or an empty array (never return null)
   * @see com.jalios.jcms.Channel#getTypeFieldEntries(Class)
   * @since jcms-5.5
   */
  public static TypeFieldEntry[] getTypeFieldsEntries() {
    return channel.getTypeFieldEntries(Publication.class);
  }

  /**
   * Return an array of <code>TypeFieldEntry</code> bound to <code>Data</code>. <br>
   * This array <b>does</b> contain inherited TypeFieldEntry from super classes of <code>Data</code>. <br>
   * <br>
   * Default implementation for <code>Data</code> will propably return an empty
   * array as no TypeFieldEntry is available for <code>Data</code>, this method
   * is available mainly to be overriden by subclasses. <br>
   * 
   * @return an array of <code>TypeFieldEntry</code> containing all Field used by <code>Data</code> or an empty array (never return null)
   * @see com.jalios.jcms.Channel#getAllTypeFieldEntries(Class)
   * @since jcms-5.5
   */
  public static TypeFieldEntry[] getAllTypeFieldEntry() {
    return channel.getAllTypeFieldEntries(Publication.class);
  }

  // ----------------------------------------------------------------------
  // FIELDs VALUE
  // ----------------------------------------------------------------------

  /**
   * Gets the value of the given <code>int</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>int</code> field value
   * @throws NoSuchFieldException if the field was not found.
   */
  public int getIntFieldValue(String fieldName) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find int field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Sets the value of the given <code>int</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param value the <code>int</code> field value
   * @throws NoSuchFieldException if the field was not found.
   * @since jcms-7.0.0
   */
  public void setIntFieldValue(String fieldName, int value) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find int field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Gets the value of the given <code>long</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>long</code> field value
   * @throws NoSuchFieldException if the field was not found.
   */
  public long getLongFieldValue(String fieldName) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find long field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Sets the value of the given <code>long</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param value the <code>long</code> field value
   * @throws NoSuchFieldException if the field was not found.
   * @since jcms-7.0.0
   */
  public void setLongFieldValue(String fieldName, long value) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find long field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Gets the value of the given <code>double</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>double</code> field value
   * @throws NoSuchFieldException if the field was not found.
   */
  public double getDoubleFieldValue(String fieldName) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find double field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Sets the value of the given <code>double</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param value the <code>double</code> field value
   * @throws NoSuchFieldException if the field was not found.
   * @since jcms-7.0.0
   */
  public void setDoubleFieldValue(String fieldName, double value) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find double field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Gets the value of the given <code>boolean</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>boolean</code> field value
   * @throws NoSuchFieldException if the field was not found.
   */
  public boolean getBooleanFieldValue(String fieldName) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find boolean field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Sets the value of the given <code>boolean</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param value the <code>boolean</code> field value
   * @throws NoSuchFieldException if the field was not found.
   * @since jcms-7.0.0
   */
  public void setBooleanFieldValue(String fieldName, boolean value) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find boolean field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Gets the value of the given <code>Category</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param mbr the member used to check rigth access to categories.
   * @return a <code>TreeSet</code> of <code>Category</code>
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   */
  public TreeSet getCategoryFieldValue(String fieldName, Member mbr) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find Category field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Sets the value of the given <code>Category</code> field name for the current <code>Data</code>.
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param value <code>TreeSet</code> of <code>Category</code>
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @since jcms-7.0.0
   */
  public void setCategoryFieldValue(String fieldName, TreeSet value) throws NoSuchFieldException {
    throw new NoSuchFieldException("Could not find Category field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Gets the value of the given <code>Object</code> field name for the current <code>Data</code>. <br>
   * If the field is multilingual, the default language will be used. If the field value
   * is not available (empty) in the given language, the field value in the main language
   * of this publication is return.<br>
   * Do not retrieve <code>Category</code> fields, see #getCategoryFieldValue(Member).
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>Object</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see #getFieldValue(String, String, boolean)
   */
  public Object getFieldValue(String fieldName) throws NoSuchFieldException {
    return getFieldValue(fieldName, channel.getLanguage(), true);
  }

  /**
   * Sets the value of the given <code>Object</code> field name for the current <code>Data</code>. <br>
   * If the field is multilingual, the default language will be used.<br>
   * Do not set <code>Category</code> fields, see #setCategoryFieldValue(String, TreeSet).
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param value the <code>Object</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see #setFieldValue(String, Object, String)
   * @since jcms-7.0.0
   */
  public void setFieldValue(String fieldName, Object value) throws NoSuchFieldException {
    setFieldValue(fieldName, value, channel.getLanguage());
  }

  /**
   * Gets the value of the given <code>Object</code> field name for the current <code>Data</code>. <br>
   * If the field is multilingual, the given language will be used. If the field value
   * is not available (empty) in the given language, the field value in the main language
   * of this publication is return.<br>
   * Do not retrieve <code>Category</code> fields, see #getCategoryFieldValue(String, Member).
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param lang the language (ISO-639 code) in which to retrieve the field value (used only for multilingual fields).
   * @return the <code>Object</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see #getFieldValue(String, String, boolean)
   */
  public Object getFieldValue(String fieldName, String lang) throws NoSuchFieldException {
    return getFieldValue(fieldName, lang, true);
  }

  /**
   * Gets the value of the given <code>Object</code> field name for the current <code>Data</code>. <br>
   * If the field is multilingual, the given language will be used. <br>
   * Do not retrieve <code>Category</code> fields, see #getCategoryFieldValue(String, Member).
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param lang the language (ISO-639 code) in which to retrieve the field value
   * (used only for multilingual fields).
   * @param useDefault whether to use the publication main language if the field value
   * is not available in the requested language (used only for multilingual fields).
   * @return the <code>Object</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   */
  public Object getFieldValue(String fieldName, String lang, boolean useDefault) throws NoSuchFieldException {
    if ("cdate".equals(fieldName)) {
      return getCdate();
    }
    if ("mdate".equals(fieldName)) {
      return getMdate();
    }
    throw new NoSuchFieldException("Could not find Object field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Sets the value of the given <code>Object</code> field name for the current <code>Data</code>. <br>
   * If the field is multilingual, the given language will be used.<br>
   * Do not set <code>Category</code> fields, see #getCategoryFieldValue(String, Member).
   * 
   * @param fieldName the field name from which to retrieve the field value.
   * @param value the <code>Object</code> field value
   * @param lang the language (ISO-639 code) in which to retrieve the field value (used only for multilingual fields).
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @since jcms-7.0.0
   */
  public void setFieldValue(String fieldName, Object value, String lang) throws NoSuchFieldException {
    if ("cdate".equals(fieldName)) {
      setCdate((Date) value);
      return;
    }
    if ("mdate".equals(fieldName)) {
      setMdate((Date) value);
      return;
    }
    throw new NoSuchFieldException("Could not find Object field '" + fieldName + "' for " + this.getClass());
  }

  /**
   * Gets the value of the given <code>Object</code> field name for this <code>Data</code>. <br>
   * Do not retrieve <code>Category</code> fields and primitive fields values.
   * 
   * @param ctxt the current context in which this field's value is request.
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>Object</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see PortalPolicyFilter#getFieldValue(JcmsJspContext, String, Data, Object)
   */
  public Object getFieldValue(JcmsJspContext ctxt, String fieldName) throws NoSuchFieldException {

    // Check parameters
    if (ctxt == null || Util.isEmpty(fieldName)) {
      return getFieldValue(fieldName);
    }

    Object value = null;
    String classShortName = Util.getClassShortName(getClass());

    String extraValueTmp;

    // Check on the portal
    extraValueTmp = getFieldValueInPortalExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = extraValueTmp;
    }

    // Check on the Member
    extraValueTmp = getFieldValueInMemberExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = extraValueTmp;
    }

    // Check on the policy filter
    value = channel.getPolicyManager().getFieldValue(ctxt, fieldName, this, value);

    // Check to return default value
    if (Util.isEmpty(value)) {
      return getFieldValue(fieldName);
    }

    // Check return type from TypeFieldEntry to cast in Data
    TypeFieldEntry tfe = channel.getTypeFieldEntry(getClass(), fieldName);
    if (tfe != null && value instanceof String && (!"String".equals(tfe.getType()))) {
      try {
        Class clazz = channel.getClass(tfe.getType());
        if (Storable.class.isAssignableFrom(clazz)) {
          return channel.getData((String) value);
        }
      } catch (Exception ex) {
        logger.warn("Can't resolve field's return type " + ex.getMessage(), ex);
      }
    }

    return value;
  }

  /**
   * Gets the integer value of the specified field name for this <code>Data</code>.
   * 
   * @param ctxt the current context in which this field's value is request.
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>int</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see PortalPolicyFilter#getFieldValue(JcmsJspContext, String, Data, Object)
   * @since jcms-6.0.0
   */
  public int getIntFieldValue(JcmsJspContext ctxt, String fieldName) throws NoSuchFieldException {

    // Check parameters
    if (ctxt == null || Util.isEmpty(fieldName)) {
      return getIntFieldValue(fieldName);
    }

    int value = getIntFieldValue(fieldName);
    String classShortName = Util.getClassShortName(getClass());

    String extraValueTmp;

    // Check on the portal
    extraValueTmp = getFieldValueInPortalExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = Util.toInt(extraValueTmp, value);
    }

    // Check on the Member
    extraValueTmp = getFieldValueInMemberExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = Util.toInt(extraValueTmp, value);
    }

    // Check on the policy filter
    value = channel.getPolicyManager().getIntFieldValue(ctxt, fieldName, this, value);

    return value;
  }

  /**
   * Gets the long value of the specified field name for this <code>Data</code>.
   * 
   * @param ctxt the current context in which this field's value is request.
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>long</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see PortalPolicyFilter#getLongFieldValue(JcmsJspContext, String, Data, long)
   * @since jcms-6.0.2, jcms-6.1.1
   */
  public long getLongFieldValue(JcmsJspContext ctxt, String fieldName) throws NoSuchFieldException {

    // Check parameters
    if (ctxt == null || Util.isEmpty(fieldName)) {
      return getLongFieldValue(fieldName);
    }

    long value = getLongFieldValue(fieldName);
    String classShortName = Util.getClassShortName(getClass());

    String extraValueTmp;

    // Check on the portal
    extraValueTmp = getFieldValueInPortalExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = Util.toLong(extraValueTmp, value);
    }

    // Check on the Member
    extraValueTmp = getFieldValueInMemberExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = Util.toLong(extraValueTmp, value);
    }

    // Check on the policy filter
    value = channel.getPolicyManager().getLongFieldValue(ctxt, fieldName, this, value);

    return value;
  }

  /**
   * Gets the double value of the specified field name for this <code>Data</code>.
   * 
   * @param ctxt the current context in which this field's value is request.
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>double</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see PortalPolicyFilter#getDoubleFieldValue(JcmsJspContext, String, Data, double)
   * @since jcms-6.0.2, jcms-6.1.1
   */
  public double getDoubleFieldValue(JcmsJspContext ctxt, String fieldName) throws NoSuchFieldException {

    // Check parameters
    if (ctxt == null || Util.isEmpty(fieldName)) {
      return getDoubleFieldValue(fieldName);
    }

    double value = getDoubleFieldValue(fieldName);
    String classShortName = Util.getClassShortName(getClass());

    String extraValueTmp;

    // Check on the portal
    extraValueTmp = getFieldValueInPortalExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = Util.toDouble(extraValueTmp, value);
    }

    // Check on the Member
    extraValueTmp = getFieldValueInMemberExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = Util.toDouble(extraValueTmp, value);
    }

    // Check on the policy filter
    value = channel.getPolicyManager().getDoubleFieldValue(ctxt, fieldName, this, value);

    return value;
  }

  /**
   * Gets the boolean value of the specified field name for this <code>Data</code>.
   * 
   * @param ctxt the current context in which this field's value is request.
   * @param fieldName the field name from which to retrieve the field value.
   * @return the <code>boolean</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see PortalPolicyFilter#getBooleanFieldValue(JcmsJspContext, String, Data, boolean)
   * @since jcms-6.0.2, jcms-6.1.1
   */
  public boolean getBooleanFieldValue(JcmsJspContext ctxt, String fieldName) throws NoSuchFieldException {

    // Check parameters
    if (ctxt == null || Util.isEmpty(fieldName)) {
      return getBooleanFieldValue(fieldName);
    }

    boolean value = getBooleanFieldValue(fieldName);
    String classShortName = Util.getClassShortName(getClass());

    String extraValueTmp;

    // Check on the portal
    extraValueTmp = getFieldValueInPortalExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = Util.toBoolean(extraValueTmp, value);
    }

    // Check on the Member
    extraValueTmp = getFieldValueInMemberExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      value = Util.toBoolean(extraValueTmp, value);
    }

    // Check on the policy filter
    value = channel.getPolicyManager().getBooleanFieldValue(ctxt, fieldName, this, value);

    return value;
  }

  /**
   * Gets the Category value of the specified field name for this <code>Data</code>.
   * 
   * @param ctxt the current context in which this field's value is request.
   * @param fieldName the field name from which to retrieve the field value.
   * @return a TreeSet of <code>Category</code> field value
   * @throws NoSuchFieldException if the field was not found in the given <code>Data</code>.
   * @see PortalPolicyFilter#getCategoryFieldValue(JcmsJspContext, String, Data, TreeSet)
   * @since jcms-6.0.2, jcms-6.1.1
   */
  public TreeSet<Category> getCategoryFieldValue(JcmsJspContext ctxt, String fieldName) throws NoSuchFieldException {

    // Check parameters
    if (ctxt == null || Util.isEmpty(fieldName)) {
      return getCategoryFieldValue(fieldName, null);
    }

    TreeSet<Category> value = getCategoryFieldValue(fieldName, ctxt.getLoggedMember());
    String classShortName = Util.getClassShortName(getClass());

    String extraValueTmp;

    // Check on the portal
    extraValueTmp = getFieldValueInPortalExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      List<Category> catList = JcmsUtil.stringToDataList(extraValueTmp, ",", Category.class);
      if (Util.notEmpty(catList)) {
        value = new TreeSet<Category>(catList);
      }
    }

    // Check on the Member
    extraValueTmp = getFieldValueInMemberExtraData(ctxt, fieldName, classShortName);
    if (extraValueTmp != null) {
      List<Category> catList = JcmsUtil.stringToDataList(extraValueTmp, ",", Category.class);
      if (Util.notEmpty(catList)) {
        value = new TreeSet<Category>(catList);
      }
    }

    // Check on the policy filter
    value = channel.getPolicyManager().getCategoryFieldValue(ctxt, fieldName, this, value);

    return value;
  }

  private String getFieldValueInPortalExtraData(JcmsJspContext ctxt, String fieldName, String classShortName) {
    Data portal = (Data) ctxt.getPortal();
    if (portal == null) {
      return null;
    }
    String extrakey = "extra." + Util.getClassShortName(portal) + "." + classShortName + "." + fieldName;
    String extraValue = portal.getExtraData(extrakey);
    return extraValue;
  }

  private String getFieldValueInMemberExtraData(JcmsJspContext ctxt, String fieldName, String classShortName) {
    Data member = ctxt.getLoggedMember();
    if (member == null || (member instanceof DBMember)) {
      return null;
    }
    String extrakey = "extra.Member." + classShortName + "." + fieldName;
    String extraValue = member.getExtraData(extrakey);
    return extraValue;
  }

  /**
   * Return the URL to access this data in the front office. <br>
   * If descriptive URLs are enabled in JCMS ({@link DescriptiveURLs})
   * the url will be generated using {@link DescriptiveURLs#getDescriptiveURL(Data, Locale)} otherwise, a standard "display.jsp?id=" URL is generated. <br>
   * <br>
   * Warning: The return url may or may not contains a query string
   * 
   * @param locale the {@link Locale} to use for I18N and localized operations
   * @return a path relative to the webapp path, (no '/' at begining).
   * @since jcms-5.7.0
   * @see JcmsUtil#getDisplayUrl(Data, Locale)
   */
  public String getDisplayUrl(Locale locale) {

    Object customDisplayUrl = getExtraInfo(DISPLAY_URL_EXTRA_INFO);
    if (customDisplayUrl != null && customDisplayUrl instanceof String) {
      return (String) customDisplayUrl;
    }

    if (DescriptiveURLs.isEnabled()) {
      return DescriptiveURLs.getDescriptiveURL(this, locale);
    }

    return "display.jsp?id=" + id;
  }
  
  /**
   * Retrieve an HTML a tag suitable to display this Data 
   * @param locale the current Locale to use for I18N of the URL and other title information
   * @return an HTML fragment 
   * @since jcms-7.1.0
   */
  public String getDisplayLink(Locale locale) {
    return getDisplayLink(locale, null, null, null, null, null, null, null);
  }

  /**
   * Retrieve an HTML a tag suitable to display this Data 
   * @param locale the current Locale to use for I18N of the URL and other title information
   * @param baseURL optionnal baseURL to use in the URL    * 
   * @return an HTML fragment 
   * @since jcms-8.0.0
   */
  public String getDisplayLink(Locale locale, String baseURL) {
    return getDisplayLink(locale, baseURL, null, null, null, null, null, null);
  }

  /**
   * Retrieve an HTML a tag suitable to display this Data 
   * @param locale the current Locale to use for I18N of the URL and other title information
   * @param baseURL optionnal baseURL to use in the URL 
   * @param params optionnal additionnal params to in the URL (eg <code>foo=bar&amp;zig=42</code>)
   * @param anchor optionnal anchor to print in the URL (eg <code>#foobar</code>
   * @param title the value to use in the title attribute (highly recommended, do not encode)
   * @param css optionnal classname to use in the tag
   * @param accessibilityInfo a string indicating where to print accessibility information of FileDocument : 
   *        <code>"title"</code> to print them in title attribute, 
   *        <code>"content"</code> to print them in the body of the a tag,
   *        anything else will not print additionnal accessibility information  
   * @param htmlAttributes optionnal additionnal html attributes to use in the tag (must no be href, title and class)
   * @return an HTML fragment 
   * @since jcms-7.1.0
   */
  public String getDisplayLink(Locale locale, String baseURL, String params, String anchor, String title, String accessibilityInfo, String css, String htmlAttributes) {
    StringBuilder sb = new StringBuilder();
    
    printDisplayLinkStart(sb, locale, baseURL, params, anchor, title, accessibilityInfo, css, htmlAttributes, null, false, null, null, null);
    printDataName(sb, locale, accessibilityInfo);
    printDisplayLinkEnd(sb);

    return sb.toString();
  }
  
  /**
   * Print opening tag &lt;a... href attribute with URL and other attributes.
   * @param sb the StringBuilder in which start of a tag will be appended
   * @param locale the user Locale
   * @param baseURL optionnal baseURL to use in the URL 
   * @param params optionnal additionnal params to in the URL (eg <code>foo=bar&amp;zig=42</code>)
   * @param anchor optionnal anchor to print in the URL (eg <code>#foobar</code>
   * @param title the value to use in the title attribute (highly recommended, do not encode)
   * @param accessibilityInfo a string indicating where to print accessibility information of FileDocument : 
   *        <code>"title"</code> to print them in title attribute, 
   *        anything else will not print additionnal accessibility information  
   * @param css optionnal classname to use in the tag
   * @param htmlAttributes optionnal additionnal html attributes to use in the tag (must no be href, title and class)
   * @param request the HttpRequest. Requested if update or paramNames is provided.
   * @param update specify this parameter to true to generate a display URL for the given data while keeping all the current others parameters.
   * @param paramNames Optionnal request parameters names to be updated from the current request parameters. You must specify the 'values' attributes with an array of the same dimension. Trigger the same url generation behavior as when specifying the 'update' attribute to true.
   * @param paramValues Request parameters values to be updated from the current request parameters. This array must match the correponding names array specified in the 'names' attribute. Trigger the same url generation behavior as when specifying the 'update' attribute to true.
   * @param paramRemove Request parameters names to be removed from the current request parameters. Trigger the same url generation behavior as when specifying the 'update' attribute to true. 
   * @since jcms-7.1.0 (was in LinkTag before)
   */
  public void printDisplayLinkStart(StringBuilder sb, Locale locale, String baseURL, String params, String anchor, String title, String accessibilityInfo, String css, String htmlAttributes, HttpServletRequest request, boolean update, String[] paramNames, String[] paramValues, String[] paramRemove) {
    sb.append("<a href=\"");
    printDisplayUrl(sb, locale, baseURL, params, anchor, request, update, paramNames, paramValues, paramRemove);
    sb.append("\"");

    if (title != null) {
      sb.append(" title=\"");
      sb.append(HttpUtil.encodeForHTMLAttribute(title));
      sb.append("\"");
    } else if (this instanceof FileDocument && "title".equals(accessibilityInfo)) {
      sb.append(" title=\"");
      sb.append(HttpUtil.encodeForHTMLAttribute(
            getDataName(I18nUtil.getLanguageKey(locale)) +
            " (" + ((FileDocument) this).getAccessibilityInformation(locale, false) + ")"));
      sb.append("\"");
    }
    
    sb.append(" class=\"");
    sb.append("ID_" + getId()); // Add DataId
    
    // Add Tooltip Card
    if ((this instanceof Content || this instanceof UserContent || this instanceof Member)
        && TTCardManager.getInstance().canAccessTTCard(channel.getCurrentLoggedMember(), this)) {
      sb.append(" ctxTooltipCard"); 
    }
    
    if (Util.notEmpty(css)){        // Add custom CSS
      sb.append(" "+css);
    }
    sb.append("\"");
    
    if (Util.notEmpty(htmlAttributes)) {
      sb.append(" ");
      sb.append(htmlAttributes);
    }
    sb.append(">");

  }

  /**
   * Print the data name in the specified StringBuilder.
   * @param sb the StringBuilder
   * @param locale the user Locale
   * @param accessibilityInfo if "content" print accessibility info for FileDocument
   * @since jcms-7.1.0 (was in LinkTag before)
   */
  public void printDataName(StringBuilder sb, Locale locale, String accessibilityInfo) {
    sb.append(getDataName(I18nUtil.getLanguageKey(locale)));
    if (this instanceof FileDocument && "content".equals(accessibilityInfo)) {
      sb.append(" (" + ((FileDocument)this).getAccessibilityInformation(locale, true) + ")");
    }   
  }
  
  /**
   * Print closing tag &lt;/a&gt;.
   * @param sb the StringBuilder
   * @since jcms-7.1.0 (was in LinkTag before)
   */
  public void printDisplayLinkEnd(StringBuilder sb) {
    sb.append("</a>");
  }

  /**
   * Prints URL using parameter of this tag.
   * @param sb the StringBuilder
   * @param locale the user Locale
   * @param baseURL optionnal baseURL to use in the URL 
   * @param params optionnal additionnal params to in the URL (eg <code>foo=bar&amp;zig=42</code>)
   * @param anchor optionnal anchor to print in the URL (eg <code>#foobar</code>
   * @since jcms-7.1.0 (was in LinkTag before)
   */
  public void printDisplayUrl(StringBuilder sb, Locale locale, String baseURL, String params, String anchor) {
    printDisplayUrl(sb, locale, baseURL, params, anchor, null, false, null, null, null);
  }
  
  /**
   * Prints URL using parameter of this tag.
   * @param sb the StringBuilder
   * @param locale the user Locale
   * @param baseURL optionnal baseURL to use in the URL 
   * @param params optionnal additionnal params to in the URL (eg <code>foo=bar&amp;zig=42</code>)
   * @param anchor optionnal anchor to print in the URL (eg <code>#foobar</code>
   * @param request the HttpRequest. Requested if update or paramNames is provided.
   * @param update specify this parameter to true to generate a display URL for the given data while keeping all the current others parameters.
   * @param paramNames Optionnal request parameters names to be updated from the current request parameters. You must specify the 'values' attributes with an array of the same dimension. Trigger the same url generation behavior as when specifying the 'update' attribute to true.
   * @param paramValues Request parameters values to be updated from the current request parameters. This array must match the correponding names array specified in the 'names' attribute. Trigger the same url generation behavior as when specifying the 'update' attribute to true.
   * @param paramRemove Request parameters names to be removed from the current request parameters. Trigger the same url generation behavior as when specifying the 'update' attribute to true. 
   * @since jcms-7.1.0 (was in LinkTag before)
   */
  public void printDisplayUrl(StringBuilder sb, Locale locale, String baseURL, String params, String anchor, HttpServletRequest request, boolean update, String[] paramNames, String[] paramValues, String[] paramRemove) {

    // Print optionnal baseURL
    if (baseURL != null) {
      sb.append(baseURL);
    }
    
    // Compute and print URL
    String url = null;
    
    if (update == true || paramNames != null || paramValues != null || paramRemove != null) {
      url = JcmsUtil.getDisplayUrlWithUpdatedParams(this, locale,
                                                    request, paramNames,
                                                    paramValues, paramRemove,
                                                    true);
    } else {
      url = getDisplayUrl(locale);
    }
    
    sb.append(url);
    
    // Additionnal params
    if (Util.notEmpty(params)) {
      // 1. no querystring at all (no '?')
      if (url.indexOf('?') == -1)  { sb.append('?'); }
      // 2. empty querystring (just the ending '?')
      else if (url.endsWith("?"))  { /* empty */ }
      // 3. querystring already as params ('?' but not at the end)
      else                         { sb.append("&amp;"); }

      // remove & or '&amp;' from given params
      if (params.startsWith("&amp;")) params = params.substring(5);
      else if (params.startsWith("&")) params = params.substring(1);
      // print params
      sb.append(params);
    }
    
    // Anchor
    if (Util.notEmpty(anchor)) {
      if (!anchor.startsWith("#")) {
        sb.append("#");
      }
      sb.append(anchor);
    }
  }

  // ----------------------------------------------------------------------
  // ICONS
  // ----------------------------------------------------------------------  
  
  /**
   * Returns source information for the icon associated to this Data, either a relative path or sprite classes name.<p>
   * 
   * If this method behavior has not been overriden by subclass, the default icon can be declared 
   * using property <code>types.{Type}.icon</code>, where <code>{Type}</code> is the short class name of
   * the type concerned.<br>
   * <ul>
   *   <li> Either by specifying a relative image path : <br/>  
   *        <pre>types.Article.icon: images/jalios/icon/abstract.gif </pre></li>
   *   <li> or else by using a sprite icon : <br/>  
   *        <pre>types.Article.icon: ss_sprite ss_date</pre></li>
   * </ul>
   * 
   * @return a relative path (eg <code>images/jalios/icon/abstract.gif</code>) or sprite classes name (eg <code>ss_sprite ss_date</code>).
   * @since jcms-8.0.0
   * @see #getDataIconHtml(String)
   */
  public String getDataIcon(){    
    String icon = channel.getProperty(JcmsConstants.TYPES_PREFIX_PROP + "" + getClass().getSimpleName() + "" + JcmsConstants.TYPES_ICON_SUFFIX_PROP, "");
    
    icon = channel.getPolicyManager().getDataIcon(this,icon);
    
    if(Util.notEmpty(icon)){
      return icon;
    }      
    
    return channel.getProperty(JcmsConstants.TYPES_PREFIX_PROP + "Data" + JcmsConstants.TYPES_ICON_SUFFIX_PROP);
  }
  
  /**
   * Returns the alternative text (alt attribute) to use in the HTML IMG tag of this Data's icon.<p>
   * 
   * If this method behavior has not been overriden by subclass, the default alternative text can be declared 
   * using language property <code>types.{Type}.alt</code>, where <code>{Type}</code> is the short class name of
   * the type concerned.<p>
   * For example :
   * <ul>
   *   <li><code>[en.prop]</code> <code>types.Workspace.alt: Workspace</code></li>
   *   <li><code>[fr.prop]</code> <code>types.Workspace.alt: Espace de travail</code></li>
   * </ul>
   * 
   * @param lang the user language
   * @return the alt Html attribute for the IMG Html tag, or empty string if non alternative text was define in property.
   * @since jcms-8.0.0
   * @see #getDataIconHtml(String)
   */
  public String getDataIconAlt(String lang){    
    String altProp = channel.getProperty(JcmsConstants.TYPES_PREFIX_PROP + "" + getClass().getSimpleName() + "" + JcmsConstants.TYPES_ICON_SUFFIX_PROP + "" + JcmsConstants.TYPES_ICON_ALT_PROP, "");
    
    altProp = channel.getPolicyManager().getDataIconAlt(this,altProp,lang);
    
    if(Util.notEmpty(altProp)){
      return JcmsUtil.glp(lang, altProp);
    }      
    
    return "";
  }
  
  /**
   * Returns the icon's tooltip. <p>
   * 
   * If this method behavior has not been overriden by subclass, the default alternative text can be declared 
   * using language property <code>types.{Type}.title</code>, where <code>{Type}</code> is the short class name of
   * the type concerned.<p>
   * 
   * @param lang the user language.
   * @return the title Html attribute for the IMG Html tag
   * @since jcms-8.0.0
   * @see #getDataIconHtml(String)
   */
  public String getDataIconTitle(String lang) {
    String titleProp = channel.getProperty(JcmsConstants.TYPES_PREFIX_PROP + "" + getClass().getSimpleName() + "" + JcmsConstants.TYPES_ICON_SUFFIX_PROP  + "" + JcmsConstants.TYPES_ICON_TITLE_PROP, "");
    
    titleProp = channel.getPolicyManager().getDataIconTitle(this,titleProp,lang);
    
    if(Util.notEmpty(titleProp)){
      return JcmsUtil.glp(lang, titleProp);
    }      
    
    return getDataName(lang);
  }
  
  /**
   * Returns the CSS for the data icon.
   * @return the CSS for the data icon.
   * @since jcms-9.0.0
   */
  public String getDataIconCSS(){
    return null;
  }

  
  /**
   * Returns the HTML code of the icon for this data in the current user language.<p>
   * 
   * This method uses value return by methods {@link #getDataIcon()}, 
   * {@link #getDataIconAlt(String)} and {@link #getDataIconTitle(String)} to build
   * an HTML IMG tag. 
   * 
   * @return the HTML code of the icon for this data.
   * @since jcms-8.0.0
   * @see Channel#getCurrentUserLang()
   * @see #getDataIcon() 
   * @see #getDataIconTitle(String) 
   * @see #getDataIconAlt(String)
   */
  public String getDataIconHtml() {
    return getDataIconHtml(channel.getCurrentUserLang());
  } 
  
  /**
   * Returns the HTML code of the icon for this data in the specified language.<p>
   * 
   * This method uses value returned by methods {@link #getDataIcon()}, 
   * {@link #getDataIconAlt(String)} and {@link #getDataIconTitle(String)} to build
   * an HTML IMG tag. 
   * 
   * @param lang the user language.
   * @return the HTML code of the icon for this data.
   * @since jcms-8.0.0
   * @see #getDataIcon() 
   * @see #getDataIconTitle(String) 
   * @see #getDataIconAlt(String)
   */
  public String getDataIconHtml(String lang) {
    if (Util.isEmpty(lang)) {
      lang = channel.getCurrentUserLang();
    }
    
    String dataIconSrc = getDataIcon();
    
    if (Util.isEmpty(dataIconSrc)) {
      return "";
    }
    
    String html = JcmsUtil.getHtmlIcon(dataIconSrc, getDataIconAlt(lang), getDataIconTitle(lang), null);
    html = channel.getPolicyManager().getDataIconHtml(this,html,lang);
    return html;
  }
  
 
  // ----------------------------------------------------------------------
  // HIBERNATE
  // ----------------------------------------------------------------------

  /**
   * Returns the value of rowId (DBData ID).
   * 
   * @return the value of rowId (DBData ID).
   * @since jcms-6.0.0
   */
  public Long getRowId() {
    return rowId;
  }

  /**
   * Sets the value of rowId (DBData ID).
   * 
   * @param rowId the rowId
   * @since jcms-6.0.0
   */
  public void setRowId(Long rowId) {
    this.rowId = rowId;

    // Fix bug JCMS-3804 : ExtraInfo entries added prior invocation of performCreate() are lost after creation
    // --> remove/restore ExtraInfoMap if 'id' is changed (as it is used for the hashcode, thus in key of DataIndexManager map)
    Map<String, Object> eiMap = DataIndexManager.getInstance().removeExtraInfoMap(this);

    this.id = HibernateUtil.buildJcmsId(getClass(), rowId);

    DataIndexManager.getInstance().setExtraInfoMap(this, eiMap);
  }

  /**
   * Returns the JStore ID of the author (only for DBData)
   * 
   * @return the JStore ID of the author (only for DBData)
   * @since jcms-6.0.0
   */
  public String getAuthorId() {
    if (Util.notEmpty(authorDBID)) {
      return authorDBID;
    }
    if (author == null) {
      return authorId;
    }
    return author.getId();
  }

  /**
   * Sets the JStore ID of the author (only for DBData).
   * <p>
   * If the specified authorId is an invalid ID (non existing Member), the author will be set to the Member specified by property <code>channel.default-author</code> (or left null if this property has not been set).
   * 
   * @param authorId the JStore author id
   * @since jcms-6.0.0
   */
  public void setAuthorId(String authorId) {
    this.authorId = authorId;
    author = channel.getMember(authorId);
    // Fix bug JCMS-1802
    if (author == null && authorId == null) {
      Member defaultAuthor = channel.getMember(channel.getProperty("channel.default-author"));
      author = defaultAuthor;
    }
  }

  /**
   * Returns true if this objet is an instance of DBData.
   * 
   * @return true if this objet is an instance of DBData.
   * @since jcms-6.2.0
   */
  public boolean isDBData() {
    return this instanceof DBData;
  }

  /**
   * Returns true if this data is stored in the database (false if it is stored in JStore).
   * 
   * @return true if this data is stored in the database (false if it is stored in JStore).
   * @since jcms-6.0.0
   */
  public boolean isInDatabase() {
    return rowId != null;
  }

  /**
   * Returns true if this data has been persisted (either in JStore or in the database).
   * 
   * @return true if this data has been persisted (either in JStore or in the database).
   * @since jcms-6.0.0
   */
  public boolean isPersisted() {
    return isStored() || isInDatabase();
  }

  @Override
  public void setId(String id) {
    // Fix bug JCMS-3804 : ExtraInfo entries added prior invocation of performCreate() are lost after creation
    // --> remove/restore ExtraInfoMap if 'id' is changed (as it is used for the hashcode, thus in key of DataIndexManager map) 
    Map<String,Object> eiMap = DataIndexManager.getInstance().removeExtraInfoMap(this);
    
    super.setId(id);
    
    DataIndexManager.getInstance().setExtraInfoMap(this, eiMap);
  }

  @Override
  public int hashCode() {
    if (getId() == null) {
      return super.hashCode();
    }
    return getId().hashCode();
  }

  @Override
  public Object clone() {
    /*
     * if (isDBData()) {
     * // log a new exception, without throwing it, in order to display stack trace of wrong code
     * Exception ex = new UnsupportedOperationException();
     * logger.error("clone() must not be invoked on Hibernate/DBData objects (" + getClass().getName() + " : " + getId() + "). *If clone() was used for an update, save will not be performed*, use getUpdateInstance() instead of clone()", ex);
     * }
     */
    return super.clone();
  }

  /**
   * Returns the merge from the given array of storable with the given array of DBID.
   * 
   * @param <T> Any Data type
   * @param clazz the class of the data contains in the returned array.
   * @param storableArray the array of storable data
   * @param dbidArray the array of DBID
   * @return the merge from the given array of storable with the given array of DBID.
   * @since jcms-6.1.0
   */
  protected <T extends Data> T[] mergeHybridLink(Class<T> clazz, T[] storableArray, String[] dbidArray) {

    if (Util.isEmpty(storableArray)) {
      return JcmsUtil.stringArrayToDataArray(clazz, dbidArray);
    }

    int length = Math.max(Util.getSize(dbidArray), Util.getSize(storableArray));
    T[] array = (T[]) Array.newInstance(clazz, length);
    System.arraycopy(storableArray, 0, array, 0, storableArray.length);
    for (int i = 0; i < dbidArray.length; i++) {
      T data = channel.getData(clazz, dbidArray[i]);
      if (data == null) {
        continue;
      }
      array[i] = data;
    }
    return array;
  }

  /**
   * Extract the DBData from the given array and return an array of identifier of those DBData.
   * The given array is cleaned.
   * 
   * @param array the array to clean.
   * @return an array of DBID
   * @since jcms-6.1.0
   */
  protected String[] computeDBIDArray(Data[] array) {
    if (array == null) {
      return null;
    }
    String[] tmp = new String[array.length];

    boolean dbdataFound = false;
    for (int i = 0; i < array.length; i++) {
      if (array[i] instanceof DBData) {
        tmp[i] = array[i].getId();
        array[i] = null;
        dbdataFound = true;
      }
    }

    if (dbdataFound) {
      return tmp;
    }
    return null;
  }

  /**
   * This methods force the collections of the DBData to be initialized()
   * 
   * @since jcms-6.2.0
   */
  public void initializeDBDataCollections() {
    if (!isDBData()) {
      throw new UnsupportedOperationException("This method is reserved for DBData");
    }
  }
  
}
