package com.jalios.jcms.handler;

import java.io.IOException;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.restlet.data.Status;

import com.jalios.jcms.ControllerStatus;
import com.jalios.jcms.Data;
import com.jalios.jcms.HttpUtil;
import com.jalios.jcms.JcmsConstants;
import com.jalios.jcms.JcmsUtil;
import com.jalios.jcms.analytics.AnalyticsManager;
import com.jalios.jcms.analytics.EventData;
import com.jalios.jcms.context.JcmsContext;
import com.jalios.jcms.context.JcmsJspContext;
import com.jalios.jcms.context.JcmsMessage;
import com.jalios.jcms.context.JcmsMessage.Level;
import com.jalios.jcms.rest.RestUtil;
import com.jalios.jcms.upload.DocUploadInfo;
import com.jalios.jcms.workspace.Workspace;
import com.jalios.util.ServletUtil;
import com.jalios.util.URLUtils;
import com.jalios.util.Util;

/**
 * This abstract class is the super class of all the <i>Form
 * Handler</i> classes (i.e. JavaBean classes used in JSP pages for
 * processing form).
 * 
 * <p>
 * To implement a custom form handler : 
 * <ul>
 *    <li>Add setters to trigger custom operations,</li>
 *    <li>Add setters for receive optionnal parameters,</li>
 *    <li>Implement {@link #processAction()} to check for any operation, validate precondition and perform action,</li>
 *    <li>Create a JSP :
 *      <ul>
 *        <li>with the handler as a JavaBean</li>
 *        <li>for which you'll invoke the validate() method</li>
 *      </ul>
 *    </li>
 * </ul>
 * 
 * <p>The following example demonstrate uses of a form handler to receive a firstname
 * as an input and when validated display a Hello World message to the user</p>  
 * 
 * Java file <code>com.example.jcmsplugin.hello.HelloWorldHandler.java</code>: 
 * <pre>
 * package com.example.jcmsplugin.hello;
 * 
 * import java.io.IOException;
 * 
 * import com.jalios.jcms.HttpUtil;
 * import com.jalios.jcms.context.JcmsMessage;
 * import com.jalios.jcms.context.JcmsMessage.Level;
 * import com.jalios.jcms.handler.JcmsFormHandler;
 * import com.jalios.util.Util;
 * 
 * public class HelloWorldHandler extends JcmsFormHandler {
 * 
 *   // Possible operations
 *   private boolean opSayHello = false;
 *   private boolean opCancel = false;
 * 
 *   // Member variables
 *   private String firstname = null; 
 * 
 *   &#064;Override
 *   public boolean processAction() throws IOException {
 *     
 *     if (validateSayHello()) {
 *       return performSayHello();
 *     }
 * 
 *     if (validateCancel()) {
 *       return performCancel();
 *     }
 * 
 *     return false;
 *   }
 * 
 *   // ---------------------------------------------------------------
 *   // Say Hello
 *   // ---------------------------------------------------------------
 * 
 *   public boolean validateSayHello() {
 *     if (!opSayHello) {
 *       return false;
 *     }
 *     if (Util.isEmpty(this.firstname)) {
 *       addMsg(new JcmsMessage(Level.WARN, glp("jcmsplugin.myplugin.myhandler.missing-firsname")));
 *       return false;
 *     }
 *     return true;
 *   }
 * 
 *   public boolean performSayHello() throws IOException {
 *     // In I18N properties : 
 *     //   jcmsplugin.myplugin.myhandler.say-hello: Hello {0} !
 *     //
 *     addMsgSession(request, new JcmsMessage(Level.INFO,  glp("jcmsplugin.myplugin.myhandler.say-hello", HttpUtil.encodeForHTML(firstname))));
 *     return false;
 *   }
 * 
 *   // ---------------------------------------------------------------
 *   // Cancel
 *   // ---------------------------------------------------------------
 * 
 *   public boolean validateCancel() {
 *     return opCancel;
 *   }
 * 
 *   public boolean performCancel() throws IOException {
 *     sendRedirect(Util.notEmpty(redirect) ? redirect : "index.jsp");
 *     return true;
 *   }
 * 
 *   // ---------------------------------------------------------------
 *   // Setters
 *   // ---------------------------------------------------------------
 * 
 *   public void setOpSayHello(String opSayHello) {
 *     this.opSayHello = true;
 *   }
 * 
 *   public void setOpCancel(String opCancel) {
 *     this.opCancel = true;
 *   }
 * 
 * 
 *   public String getFirstname() {
 *     return this.firstname;
 *   }
 *   public void setFirstname(String firstname) {
 *     this.firstname = firstname;
 *   }
 * }
 * </pre>
 * JSP <code>sayHello.jsp</code>: 
 * <pre>
 * <%@ include file='/jcore/doInitPage.jspf' %>&lt;%
 * %>&lt;jsp:useBean id="formHandler" scope="page" class="com.example.jcmsplugin.hello.HelloWorldHandler">&lt;%
 *  %>&lt;jsp:setProperty name="formHandler" property="request" value="&lt;%= request %>" />&lt;%
 *  %>&lt;jsp:setProperty name="formHandler" property="response" value="&lt;%= response %>" />&lt;%
 *  %>&lt;jsp:setProperty name="formHandler" property="*" />&lt;%
 * %>&lt;/jsp:useBean>&lt;%
 * 
 *   if (formHandler.validate()) {
 *     return;
 *   }
 * 
 * 
 * %>&lt;%@ include file='/jcore/doEmptyHeader.jspf' %>&lt;%
 * %>&lt;%@ include file='/jcore/doMessageBox.jspf' %>
 * 
 * &lt;div id='hello'>
 *  &lt;form action="sayHello.jsp" name="sayHelloForm" method="post">
 *   &lt;%-- Inputs --%>
 *   &lt;jalios:field name="firstname" formHandler="&lt;%= formHandler %>">
 *     &lt;jalios:control />
 *   &lt;/jalios:field>
 *   &lt;%-- Hidden Inputs --%>
 *   &lt;% if (Util.notEmpty(formHandler.getRedirect())) { %>
 *   &lt;input type="hidden" name="redirect" value="&lt;%= encodeForHTMLAttribute(formHandler.getRedirect()) %>" />
 *   &lt;% } %>
 *   &lt;%-- Buttons --%>
 *   &lt;%= WidgetUtil.printHtmlButton("opSayHello", glp("jcmsplugin.btn.sayhello"), true, null, true) %>
 *   &lt;%= WidgetUtil.printHtmlButton("opCancel", glp("ui.com.btn.cancel"), true, null, false) %>
 *  &lt;/form>
 * &lt;/div>
 * </pre>
 * 
 * Data Handlers (handlers inheriting from {@link EditDataHandler}) must perform
 * additionnal operation, see corresponding JavaDoc. 
 *
 * @version $Revision: 95054 $
 */
public abstract class JcmsFormHandler extends JcmsJspContext implements JcmsConstants {
  public static final String REVISION = "$Revision: 95054 $";
  
  private static final Logger logger = Logger.getLogger(JcmsFormHandler.class);
  
  protected String    redirect             = "";
  protected boolean   noRedirect           = false;
  protected boolean   noSendRedirect       = false;
  protected boolean   redirectOnClosePopup = false;
  protected boolean   workspaceForced = false;
  protected boolean   popupEdition    = false;
  protected HashSet   editFieldSet    = new HashSet();
  protected HashMap<String, Object>   contextMap;
  protected int formStep;
  
  //-------------------------------------------------------------------
  // Validate and process action methods
  //-------------------------------------------------------------------

  /**
   * Method to implement if you need to initialize your handler from parameter received,
   * prior to all security validation. <p>
   * You <strong>must absolutely not</strong> perform any action which modify data or
   * site configuration.<p>
   * <strong>Therefore it is strongly recommend NOT TO implement this method, or use
   * very wisely.
   */
  protected void init() {
  }
  
  /**
   * Method invoked by JSP to validate all conditions (authorization, parameters, ...),
   * and eventualy trigger the action that was requested to the handler. 
   * <p>
   * You MUST not override this method to validate and perform your action, instead
   * implement {@link #processAction()} if you need to provide custom actions.
   * <p>
   * 
   * As of current implementation, this method is NOT final to ensure backward compatibility
   * with older handler which may have had a validate() method. But it WILL end as a final method. 
   * 
   * @return true if a redirect was done, false otherwise
   * @throws IOException if error occured during validate
   * @since jcms-7.1 in JcmsFormHandler (was in EditDataHandler), 
   */
  public /* TODO : final */ boolean validate() throws IOException {
    // throw new RuntimeException("test");

    // Pre-initialization
    init();
    
    // Precondition checks
    // Cancel any processing if an operation was requested but the CSRF validation fails
    if (HttpUtil.isCSRFEnabled()) {
      List<String> operationList = getRequestedOperations();
      if (Util.notEmpty(operationList) && !checkCSRF()) {
        if (!RestUtil.isRest(request)) {
          // sendForbidden(request, response);
          addMsg(new JcmsMessage(Level.WARN, glp("msg.security.csrf", Util.join(operationList, ","), ServletUtil.getRequestId(request))));
        } else {
          response.setStatus(Status.CLIENT_ERROR_FORBIDDEN.getCode());
        }
        // log a CSRF event
        try {
          AnalyticsManager.appendEventData(new EventData(AnalyticsManager.CORE,AnalyticsManager.CSRF_ERROR),request);
        } catch (Exception e) {
          logger.warn(e.getMessage(), e);
        }
        return false;
      }
    }
    
    // Perform action if any
    boolean result = this.processAction();

    // Terminate with proper return status code
    // (overriden by EditDataHandler to return proper HTTP status code when working with REST) 
    return afterValidation(result);
  }  
  
  /**
   * Find all operation requested to this handler and return them all in a list.
   * <p>
   * This method voluntarly exclude the 'opSearch' parameter which is not an operation
   * leading to server side modification.
   * 
   * @return a list of operation parameter, properly encoded for use in HTML attribute.
   */
  @SuppressWarnings("rawtypes")
  private List<String> getRequestedOperations() {
    List<String> operationList = new ArrayList<String>();
    for (Enumeration enu = request.getParameterNames(); enu.hasMoreElements();) {
      String paramName = (String) enu.nextElement();
      if (paramName.matches("op[A-Z].*") && 
          !"opSearch".equals(paramName)) { // Always ignore 'opSearch' param (fix bug JCMS-3157 and others...) 
        operationList.add(HttpUtil.encodeForHTML(paramName));
      }
    }
    return operationList;
  }
  
  
  protected boolean afterValidation(boolean result) throws IOException {
    return result;
  }
  
  /**
   * Method to be implemented to check/validate action to be performed and process them.
   * <p>
   * Default implementation is to return false. No need to call super method.
   * <p>
   * This method may NOT be invoked at all if security validation are not met.<br/>
   * You must ensure your handler and JSP can work properly without any code invoked
   * in this method.
   * 
   * @return false if no redirection is needed, true otherwise
   * @throws java.io.IOException IO Exception 
   */
  public boolean processAction() throws java.io.IOException {
    return false;
  }
  
  // ----------------------------------------------
  // Language
  // ----------------------------------------------

  public String getAvailableMainLanguage() {
    return channel.getLanguage();
  }
  
  // ----------------------------------------------
  // Steps Management
  // ----------------------------------------------
  
  /**
   * Check if the "Cancel" button should be displayed in step modal form.
   * <p>
   * The default implementation of this method is to always return true.<br/>
   * You can provide a custom FormHandler to override this behavior.
   * <p>
   * See jcore/doStepButtons.jspf for use of this method.
   * 
   * @return true to display the "Cancel" button, false to hide it
   * @since jcms-7.1.2 jcms-7.2 as requested by issue <a href="http://issues.jalios.com/browse/JCMS-3057">JCMS-3057</a>   
   */
  public boolean showCancelButton() {
    return true;
  }

  /**
   * Check if the "Next" button should be displayed in step modal form.
   * <p>
   * The default implementation of this method is to return true
   * unless the finish button is being displayed
   * by method {@link #showFinishButton()}.<br/>
   * You can provide a custom FormHandler to override this behavior.
   * <p>
   * See jcore/doStepButtons.jspf for use of this method.
   *  
   * @return true to display the "Next" button, false to hide it
   */
  public boolean showNextButton() {
    return !showFinishButton();
  }

  /**
   * Check if the "Previous" button should be displayed in step modal form.
   * <p>
   * The default implementation of this method is to return true
   * if the current step (as returned by {@link #getFormStep()} is higher than 0.<br/>
   * You can provide a custom FormHandler to override this behavior.
   * <p>
   * See jcore/doStepButtons.jspf for use of this method.
   *  
   * @return true to display the "Previous" button, false to hide it
   */
  public boolean showPreviousButton() {  
    return formStep > 0;
  }

  /**
   * Check if the "Finish" button should be displayed in step modal form.
   * <p>
   * The default implementation of this method is to return true
   * if the current step (as returned by {@link #getFormStep()} is 
   * the last one (as compared with {@link #getFormStepCount()}.<br/>
   * You can provide a custom FormHandler to override this behavior.
   * <p>
   * See jcore/doStepButtons.jspf for use of this method.
   *  
   * @return true to display the "Finish" button, false to hide it
   */
  public boolean showFinishButton() {
    return formStep == getFormStepCount() - 1;
  }
  
  /**
   * Retrieve the total number of steps used for edition when invoking this form handler. 
   * <p>
   * The default implementation of this method is to return 1, corresponding to one step (which means only 1 screen being displayed).<br/>
   * You can provide a custom FormHandler to override this behavior.
   * <p>
   * See jcore/doSteps.jspf for use of this method.
   * 
   * @return the total number of steps in this formhandler
   */
  public int getFormStepCount() {
    return 1;
  }
  
  
  /**
   * @return the prefix prop for the steps
   * @since jcms-9.0.0
   */
  public String getFormStepPrefixProp() {
    return "";
  }
  
  /**
   * @return the label of the Finish button
   * @since jcms-9.0.0
   */
  public String getFormStepFinishLabel() {
    return "ui.com.btn.finish";
  }
  
  /**
   * Returns the label array for the steps.
   * @param userLang the language
   * @return the label array for the steps.
   * @since jcms-9.0.0
   */
  public String[] getFormStepLabels(String userLang) {
    int stepCount = getFormStepCount();
    
    if (stepCount < 2) {
      return new String[]{getFormStepPrefixProp() + "1"};
    }
    
    String[] labels = new String[stepCount];
    
    for (int i = 0; i < stepCount; i++) {
      String label = JcmsUtil.glp(userLang, getFormStepPrefixProp() + i);
      labels[i] = label;
    }
    
    return labels;
  }
  
  
  public String getFormStepHiddenFields(){
    return "";
  }
  
  /**
   * Retrieve the current step of edition when invoking this form handler. 
   * <p>
   * See jcore/doSteps.jspf for use of this method.
   * 
   * @return the current steps in this formhandler (starting at 0 and ending at <code>{@link #getFormStepCount()} - 1</code>)
   */
  public int getFormStep() {
    return formStep;
  }

  /**
   * Set the current step of edition when invoking this form handler. 
   * 
   * @param step the current step (0 being the first one and <code>{@link #getFormStepCount()} - 1</code> being the last one)
   */
  public void setFormStep(int step ) {
    this.formStep = step ;
  } 
  
  // -------------------------------------------------------------------
  // Workspace getter/setter
  // -------------------------------------------------------------------

  /**
   * @return the workspace (the one set by doInitPage.jsp)
   */
  @Override
  public Workspace getWorkspace() {
    if (workspace != null || workspaceForced) {
      return workspace;
    }
    
    if (request == null) {
      return workspace;
    }
    
    workspace = super.getWorkspace();
    return workspace;
  }
  
  public boolean isWorkspaceForced(){
    return workspaceForced;
  }
  
  public void setWorkspace(Workspace wspc){
    if (logger.isDebugEnabled()) {
      logger.debug("Setting Workspace: " + wspc);
    }
    this.workspace       = wspc;
    this.workspaceForced = true;
  }
  
  public void setWs(String v){
    if (!workspaceForced){
      if (logger.isDebugEnabled()) {
        logger.debug("Setting WS: " + v);
      }
      setWorkspace(channel.getWorkspace(v)); // voluntarly invoke setWorkspace to ensure QueryHandler benefit from this value as setWorkspace(Workspace ) is overrided in QueryHandler
    }
  }
  
  // -------------------------------------------------------------------
  // Redirect method & getter/setter
  // -------------------------------------------------------------------
  
  /**
   * Set the redirect in the response. If noRedirect has been set,
   * this method set no redirection. If noSendRedirect has been set,
   * put the request attribute "redirect" but do not call
   * response.sendRedirect(). This may be useful when redirecting from
   * a dynamically included JSP in the front-office (some AppServer
   * throw an exception when setting/getting request attribute once
   * the redirect has been put.)
   *
   * @param url the url to redirect to.
   * @param checkRedirect if true and if redirect attribute has been set, redirect on this url
   * @throws java.io.IOException IO Exception 
   * @since jcms-4.0
   */
  public void sendRedirect(String url, boolean checkRedirect) throws java.io.IOException {

    // WARNING: The code of this method is copied from the static method
    //            JcmsContext#sendRedirect(String, HttpServletRequest, HttpServletResponse)
    //          and slightly modified, make sure you modify both method.
    
    if (noRedirect) {
      return;
    }

    HttpServletRequest request = getRequest();
    HttpServletResponse response = getResponse();

    String redirect = getRedirect();

    if (checkRedirect && Util.notEmpty(redirect)) {
      url = redirect;
    }
    
    // Check validity or requested redirect URL: default to webapp root if invalid 
    if (!HttpUtil.isValidHttpUrl(url)) {
      JcmsUtil.logSecurityIssue(logger, "Redirect attempt to invalid URL '"+url+"'. User will be redirected to webapp root. Update configuration if URL should be authorized (see javadoc for HttpUtil.isValidHttpUrl()).");
      url = ServletUtil.getBaseUrl(request);
    }
    
    if (!url.startsWith("http") && !url.startsWith("/")) {
      url = ServletUtil.getBaseUrl(request) + url;
    }
    
    String forceRedirect = (String)request.getAttribute(JcmsConstants.FORCE_REDIRECT);
    if (Util.notEmpty(forceRedirect)) {
      url = forceRedirect;
    }

    if (channel.isReverseProxyEnabled()) {
      url = channel.processReverseProxyRedirect(request, url);
    }    
    
    boolean ignoreUrl = url != null && (url.indexOf("makeLink2.jsp") != -1 || url.indexOf("jcore/closePopup.jsp") != -1);
    if (popupEdition && !ignoreUrl && !Util.toBoolean(request.getAttribute("ignoreClosePopup"), false)) {
      String closePopupUrl = getClosePopupUrl();
      if (redirectOnClosePopup && redirect != null) {
        closePopupUrl = URLUtils.buildUrl(closePopupUrl, Util.getHashMap("redirect", new String[] { ServletUtil.encodeURL(url) }));
      }
      url = closePopupUrl;
    }

    request.setAttribute("redirect", url); // used by doDisplayBuffer.jsp and ajaxPortal.jsp
    
    boolean isIncluded = request.getAttribute("javax.servlet.include.request_uri") != null;
    boolean isCommitted = response.isCommitted();
    boolean isRest = RestUtil.isRest(request);
    if (!noSendRedirect && !isIncluded && !isCommitted && !isRest) {
      url = ServletUtil.unescapeAmpersand(url);
      
      // Improvement JCMS-3078 : Allow configuration of HTTP status code used for redirect 
      // 1. Read default value from "channel.redirect"
      int redirectStatusCode = channel.getIntegerProperty("channel.redirect", HttpServletResponse.SC_MOVED_TEMPORARILY);
      // 2. When in Front Office, use other propery, if available
      if (Util.toBoolean(request.getAttribute("inFO"), false)) {
        redirectStatusCode = channel.getIntegerProperty("channel.redirect.front-office", redirectStatusCode);
      }
      // 3. Allow redirection to be forced by request attribute (needed by display.jsp/PortalRedirect)
      int forcedRedirecteStatusCode = Util.toInt(request.getAttribute("force-redirect-status-code"), -1);
      if (forcedRedirecteStatusCode > 0) {
        redirectStatusCode = forcedRedirecteStatusCode;
      }      
      
      // 4. Redirect
      response.setStatus(redirectStatusCode);
      response.setHeader("Location",ServletUtil.unescapeAmpersand(url)); 
    }
  }
  
  protected String getClosePopupUrl() {
    return ServletUtil.getBaseUrl(request) + "jcore/closePopup.jsp";
  }
  
  /**
   * Redirect current request to given URL.<br>
   * Overrides {@link JcmsContext#sendRedirect(String)} to use behavior
   * provided by {@link #sendRedirect(String, boolean)}.
   *
   * @param url the url to redirect to.
   * @since jcms-4.0
   */
  @Override
  public void sendRedirect(String url) throws java.io.IOException {
    sendRedirect(url, false);
  }
  
  /**
   * Get the value of redirect.
   * @return Value of redirect.
   */
  public String getRedirect() {
    return redirect;
  }
  
  /**
   * Set the value of redirect.
   * @param url  Value to assign to redirect.
   */
  public void setRedirect(String url) {
    this.redirect = HttpUtil.validateHttpUrl(url);
    if (url != null && redirect == null) {
      JcmsUtil.logSecurityIssue(logger, "Invalid 'redirect' URL parameter : '"+url+"'. Update configuration if URL should be authorized (see javadoc for HttpUtil.isValidHttpUrl()).");
    }
  }

  /**
   * Get the redirect value to be used in the modals with modalRedirect.jspf and the request attribute 'modal.redirect'.
   * 
   * @return redirect value
   * @since jcms-9.0.4
   */
  public String getModalRedirect() {
    // refresh the current page
    return null;
  }

  /**
   * Set the value of noRedirect.
   * @param noRedirect  Value to assign to noRedirect.
   * @since jcms-4.0
   */
  public void setNoRedirect(boolean  noRedirect) {
    this.noRedirect = noRedirect;
  }
  
  /**
   * Set the value of noSendRedirect.
   * @param noSendRedirect  Value to assign to noSendRedirect.
   * @since jcms-4.0
   */
  public void setNoSendRedirect(boolean  noSendRedirect) {
    this.noSendRedirect = noSendRedirect;
  }
  
  /**
   * Set whether the redirect value should be applied to the opener when
   * closing the popup edition.<br>
   * The value set by this method will be used only when using popup edition 
   * ({@link #setPopupEdition(boolean)}).
   * @param redirectOnClosePopup if true, the redirect value ({@link #setRedirect(String)})
   *        will be provided to closePopup.jsp to redirect the opener. if false,
   *        the popup will simply be closed and the opener refreshed.
   * @since jcms-5.7
   */
  public final void setRedirectOnClosePopup(boolean  redirectOnClosePopup) {
    this.redirectOnClosePopup = redirectOnClosePopup;
  }
  
  /**
   * Return the value indicating whether the redirect value should be
   * applied to the opener when closing the popup edition.<br>
   * @return true, when redirect value ({@link #setRedirect(String)})
   *        will be provided to closePopup.jsp to redirect the opener. false,
   *        if the popup will simply be closed and the opener refreshed.
   * @since jcms-5.7
   */
  public final boolean getRedirectOnClosePopup() {
    return this.redirectOnClosePopup;
  }

  // -------------------------------------------------------------------
  // Workspace getter/setter
  // -------------------------------------------------------------------
  
  /**
   * Process the ControllerStatus: do nothing if OK, or send on
   * forbidden, or put a warning message in the request.
   *
   * @param status the ControllerStatus to be processed
   * @return true when status is OK, false if status was a failure (forbidden or anything else)
   * @throws java.io.IOException if an error occurs
   * @since jcms-5.0.0
   */
  protected boolean processStatus(ControllerStatus status) throws java.io.IOException {
    
    // OK
    if (!status.hasFailed()) {
      return true;
    }
    
    // Forbidden
    if (status.isForbidden()) {
      sendForbidden();
      return false;
    }
    
    // Warning
    JcmsMessage warnMsg = new JcmsMessage(Level.WARN, status.getMessage(getUserLang()));
    addMsg(warnMsg);
    return false;
  }
  
  
  /**
   * Build a controller context. I.E. add request, response and loggedMember.
   *
   * @return the HashMap which contains the context
   * @since jcms-5.0.0
   */
  protected HashMap<String, Object> getControllerContext() {
    if (contextMap == null){
      contextMap = new HashMap<String, Object>();
      contextMap.put("request",      request);
      contextMap.put("response",     response);
      contextMap.put("loggedMember", loggedMember);
      contextMap.put("formHandler" , this);
    }
    return contextMap;
  }
  
  
  /**
   * Check if the given parameter is present in the query parameters
   *
   * @param field the field to be checked
   * @return true if this field is present in the query parameters
   * @since jcms-5.0.0
   */
  protected boolean isFieldMissing(String field) {
    return !HttpUtil.hasParameter(request, field);
    //return request.getPar ameter(field) == null;
  }
  
  /**
   * Set where this form handler is being invoked in a popup.
   * @param v true if handler is invoked in popup, false otherwise
   * @since jcms-5.0.0
   */
  public void setPopupEdition(boolean v) {
    popupEdition = v;
  }
  
  /**
   * @return true if this form handler is invoked through a popup
   * @since jcms-5.0.0
   */
  public boolean isPopupEdition() {
    return popupEdition;
  }
  
  /**
   * @param v a set of field names to edit
   * @since jcms-5.0.0
   */
  public void setEditField(String[] v) {
    for(int i = 0; i < v.length; i++) {
      editFieldSet.add(v[i]);
    }
  }
  
  /**
   * @return a set of field names being edited
   * @since jcms-5.0.0
   */
  public HashSet getEditFieldSet() {
    return editFieldSet;
  }
  
  /**
   * @param field field name
   * @return true if this handler is invoked to edit only some fields of the data
   * @since jcms-5.0.0
   * @see #getEditFieldSet()
   */
  public boolean isFieldEdition(String field) {
    if (Util.isEmpty(editFieldSet)) {
      return true;
    }
    return editFieldSet.contains(field);
  }
  
  /**
   * Returns true if this is a partial field edition (i.e. front edition)
   * 
   * @return true if this is a partial field edition.
   * @since jcms-5.5.0
   */
  public boolean isPartialFieldEdition() {
    return Util.notEmpty(editFieldSet);
  }
  
  //-----------------------------------------------
  // Upload
  //-----------------------------------------------
  
  /**
   * Perform upload then, return value of the real File realtive path.
   * if value is equal to MultipartRequest.MULTIPART_UPLOAD
   * 
   * @param fieldName the field name
   * @param value the field fake value
   * @param isFileDocument  is file document
   * @param override override file 
   * @return String 
   * @since jcms-5.7.0
   */
  protected String updateUploadedField(String fieldName, String value, boolean isFileDocument, boolean override){
    if (Util.isEmpty(value))
      return value;
    
    if (!com.jalios.jcms.upload.MultipartRequest.MULTIPART_UPLOAD.equals(value))
      return value;
    
    DocUploadInfo dui = (DocUploadInfo) getUploadedFile(fieldName,isFileDocument,override);
    return dui != null ? dui.getFilename() : "";
  }
  
  /**
   * Perform upload then, replace all values with value MultipartRequest.MULTIPART_UPLOAD
   * with the real File realtive path.
   * 
   * @param fieldName the field name
   * @param values the field fake values
   * @param isFileDocument  is file document
   * @param override override file 
   * @since jcms-5.7.0
   */
  protected void updateUploadedFields(String fieldName, String[] values, boolean isFileDocument, boolean override){
    if (Util.isEmpty(values))
      return;
    
    List<DocUploadInfo> uploadList = getUploadedFileList(fieldName,isFileDocument,override);
    if (uploadList == null) {
      return;
    }
    for (int i = 0, cpt= 0; i < values.length; i++) {
      if (!com.jalios.jcms.upload.MultipartRequest.MULTIPART_UPLOAD.equals(values[i])){
        continue;
      }
      if (uploadList.size() <= cpt+1){ 
        continue; 
      }
      DocUploadInfo dui = uploadList.get(cpt++);
      values[i] = dui != null ? dui.getFilename() : "";
    }
  }
  
  //-----------------------------------------------
  // ML
  //-----------------------------------------------
  
  // TODO javadoc
  /**
   * Returns the first value (i.e. the main language value) from the given array (monovalued ML fields).
   * 
   * @param array the array
   * @param trim if true trim the value
   * @param escape if true escape the value
   * @return the first value from the array.
   * @since jcms-6.0.0
   */
  protected String getMainLangValue(String[] array, boolean trim, boolean escape) {
    if (array == null) {
      return null;
    }
    String str = array[0];
    if (str == null) {
      str = "";
    }
    if (trim) {
      str = str.trim();
    }
    
    if (!escape) {
      return str;
    }
    
    return JcmsUtil.escapeHtml(str);
  }
  
  /**
   * Returns the ML map (i.e. the additionnal languages) filled with the values contained in the given array  (monovalued ML fields).
   * 
   * @param array the array
   * @param trim if true trim the values
   * @param escape if true escape the values
   * @return the ML map filled with the values contained in the given array.
   * @since jcms-6.0.0
   */
  protected HashMap<String,String> getMLMap(String[] array, boolean trim, boolean escape) {
    if (array == null) {
      return null;
    }
    array = Util.trim(array);
    
    if (array.length < 2) {
      return null;
    }
    HashMap<String,String> mlMap = new HashMap<String,String>();
    List<String> otherLangList = channel.getOtherLanguageList();
    for(int i=1; i < array.length && i -1 < otherLangList.size(); i++) {
      String str = array[i];
      
      // Prevent bug DEVJCMS-504 if a setter is invoked with invalid parameter array
      if (str == null) {
        str = "";
      }
      if (trim) {
        str = str.trim();
      }
      if (escape) {
        str = JcmsUtil.escapeHtml(str); 
      }
      
      String lang = otherLangList.get(i-1);
      mlMap.put(lang, str);
    }
    return mlMap;
  }
  
  /**
   * Returns the first values (i.e. the main language value) from the given array (multivalued ML fields).
   * 
   * @param array the array
   * @param trim if true trim the values
   * @param escape if true escape the values
   * @return the first values from the array (i.e. the main language value) from the given array.
   * @since jcms-6.0.0
   */
  protected String[] getMainLangValueArray(String[] array, boolean trim, boolean escape) {
  
    List<String> list = new ArrayList<String>();
    int nbItem = array.length / channel.getLanguageList().size();
    for(int i=0; i < nbItem; i++) {
      String str = array[i];
      if (str == null) {
        str = "";
      }
      if (trim) {
        str = str.trim();
      }
      if (escape) {
        str = JcmsUtil.escapeHtml(str);
      }
      list.add(str);
    }
    return Util.trimStringArray(list.toArray(new String[0]));
  }
  
  /**
   * Returns the ML map (i.e. the additionnal languages) filled with the values contained in the given array (multivalued ML fields).
   * 
   * @param array the array
   * @param trim if true trim the values
   * @param escape if true escape the values
   * @return the ML map filled with the values contained in the given array.
   * @since jcms-6.0.0
   */
  protected HashMap<String,String[]> getMLMapArray(String[] array, boolean trim, boolean escape) {
    if (array == null || Util.isEmpty(Util.join(array,""))) {
      return null;
    }
    List<String> list = new ArrayList<String>();
    int nbItem = array.length / channel.getLanguageList().size();

    HashMap<String,String[]> mlMap = new HashMap<String,String[]>();
    List<String> otherLangList = channel.getOtherLanguageList();
    for(int i=1; i < array.length && i -1 < otherLangList.size(); i++) {
      list.clear();
      for(int j=0; j < nbItem; j++) {
        String str = array[i*nbItem + j];   
        if (str == null) {
          str = "";
        }
        if (trim) {
          str = str.trim();
        }
        if (escape) {
          str = JcmsUtil.escapeHtml(str);
        }
        list.add(str);
      }
      String lang = otherLangList.get(i-1);
      String[] values = Util.trimStringArray(list.toArray(new String[0]));
      mlMap.put(lang, values);
    }
    
    return mlMap;
  }


  // Hidden fields helpers

  public String getHiddenField(String name, String value) {
    return "<input type=\"hidden\" name=\"" + name + "\" value=\"" + HttpUtil.encodeForHTMLAttribute(value, true) + "\" />\n";
  }

  public String getHiddenFieldML(String name, HashMap<String, String> map) {
    if (Util.isEmpty(map)) {
      return "";
    }
    StringBuilder sb = new StringBuilder();
    List<String> languageList = channel.getLanguageList();
    for (Iterator<String> iterator = languageList.iterator(); iterator.hasNext();) {
      String lang = iterator.next();
      String value = map.get(lang);
      sb.append(getHiddenField(name, value));
    }
    return sb.toString();
  }

  public String getHiddenField(String name, Data value) {
    if (value != null) {
      return getHiddenField(name, value.getId());
    }
    return "";
  }

  public String getHiddenField(String name, boolean value) {
    return getHiddenField(name, Boolean.toString(value));
  }

  public String getHiddenField(String name, int value) {
    return getHiddenField(name, Integer.toString(value));
  }

  public String getHiddenField(String name, double value) {
    return getHiddenField(name, Double.toString(value));
  }
  

  public String getHiddenField(String name, long value) {
    return getHiddenField(name, Long.toString(value));
  }

  public String getHiddenField(String name, Date value) {
    if (value != null) {
      String dateStr = channel.getDateTimeFormat(userLang).format(value);
      return getHiddenField(name, dateStr);
    }
    return "";
  }

  public String getHiddenField(String name, Data[] datas) {
    StringBuilder sb = new StringBuilder();
    Set<String> idSet = JcmsUtil.dataArrayToIdSet(datas);
    if (Util.notEmpty(idSet) ) {
      for (String id : idSet) {
        sb.append(getHiddenField(name, id));
      }
    }
    else{
      sb.append(getHiddenField(name, ""));
    }
    return sb.toString();
  }

  public String getHiddenField(String name, boolean[] array) {
    StringBuilder sb = new StringBuilder();
    if (Util.notEmpty(array) ) {
      for (boolean value : array) {
        sb.append(getHiddenField(name, Boolean.toString(value)));
      }
    }
    else{
      sb.append(getHiddenField(name, ""));
    }
    return sb.toString();
  }

  public String getHiddenField(String name, int[] array) {
    StringBuilder sb = new StringBuilder();
    if (Util.notEmpty(array) ) {
      for (int value : array) {
        sb.append(getHiddenField(name, Integer.toString(value)));
      }
    }
    else{
      sb.append(getHiddenField(name, ""));
    }
    return sb.toString();
  }


  public String getHiddenField(String name, long[] array) {
    StringBuilder sb = new StringBuilder();
    if (Util.notEmpty(array) ) {
      for (long value : array) {
        sb.append(getHiddenField(name, Long.toString(value)));
      }
    }
    else{
      sb.append(getHiddenField(name, ""));
    }
    return sb.toString();
  }
  
  public String getHiddenField(String name, double[] array) {
    StringBuilder sb = new StringBuilder();
    if (Util.notEmpty(array) ) {
      for (double value : array) {
        sb.append(getHiddenField(name, Double.toString(value)));
      }
    }
    else{
      sb.append(getHiddenField(name, ""));
    }
    return sb.toString();
  }

  public String getHiddenField(String name, String[] array) {
    StringBuilder sb = new StringBuilder();
    if (Util.notEmpty(array) ) {
      for (String value : array) {
        sb.append(getHiddenField(name, value));
      }
    }
    else{
      sb.append(getHiddenField(name, ""));
    }
    return sb.toString();
  }

  public String getHiddenField(String name, Date[] array) {
    if (Util.isEmpty(array)) {
      return getHiddenField(name, "");
    }
    StringBuilder sb = new StringBuilder();
    for (Date date : array) {
      String dateStr = "";
      if( date  != null){
        dateStr = channel.getDateTimeFormat(userLang).format(date);
      }
      sb.append(getHiddenField(name, dateStr));
    }
    return sb.toString();
  }

  public <T extends Data> String getHiddenField(String name, Set<T> set) {
    if (Util.notEmpty(set)) {
      Class<? extends Data> clazz = set.iterator().next().getClass();
      T[] newInstance = (T[]) Array.newInstance(clazz, set.size());
      T[] array = set.toArray(newInstance);
      return getHiddenField(name, array);
    }
    return getHiddenField(name, "");
  }
  
  protected boolean checkMissingField(Object obj, String prop) {    
    if (Util.notEmpty(obj)) {
      return true;
    }
    
    addMsg( new JcmsMessage(Level.WARN, glp("msg.edit.empty-field", glp(prop))));
    return false;
  }
}
