package com.jalios.jcms.taglib.settings.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.jalios.jcms.Category;
import com.jalios.jcms.Channel;
import com.jalios.jcms.Data;
import com.jalios.jcms.JcmsUtil;
import com.jalios.jcms.Publication;
import com.jalios.jcms.TypeFieldEntry;
import com.jalios.jcms.taglib.ControlType;
import com.jalios.jcms.taglib.settings.AbstractDataSettings;
import com.jalios.jcms.taglib.settings.ControlUtils;
import com.jalios.util.LangProperties;
import com.jalios.util.Util;

/**
 * Publication Settings
 * 
 * @author Nicolas Dos Santos
 */
public class PublicationSettings extends AbstractDataSettings<PublicationSettings> {

  private static final Logger LOG = Logger.getLogger(PublicationSettings.class);

  public static final String SUPER_TYPE = "superType";
  public static final String SUPER_TYPE_CLASS = "superTypeClass";
  public static final String PSTATUS = "pstatus";
  public static final String CATEGORIES = "categories";
  public static final String NO_WORKSPACE_FILTER = "noWorkspaceFilter";
  public static final String DOCUMENT_KINDS = "documentKinds";
  public static final String EXACT_TYPE = "exactType";
  public static final String FILE_INPUT_DISABLED = "fileInputDisabled";
  public static final String FILE_INPUT_ENABLED = "fileInputEnabled";
  public static final String DROPDOWN_DISABLED = "dropdownDisabled";

  private String rootCategoryId;
  private boolean displayRoot;

  private static List<String> documentKindsList = null;

  @Override
  public ControlType getType() {
    return ControlType.PUBLICATION;
  }

  @Override
  protected PublicationSettings getThis() {
    return this;
  }

  @Override
  public Class<? extends Data> getDataClass() {
    Class<? extends Data> superTypeClass = getOption(SUPER_TYPE_CLASS);
    return superTypeClass != null ? superTypeClass : Publication.class;
  }

  @Override
  public <V> V getOption(String optionName) {
    boolean channelReady = isChannelReady();

    // Resolve the superType Class from 'superType' if 'superType' exists and superType Class is not resolved yet
    // Check if the superType has required fields to disable the file input
    resolveSuperTypeClass(optionName, channelReady);

    /* Hack to resolve the options dependent on the Store. */
    if (channelReady) {
      /* We don't do it in resolveOptions() method because the Store is not loaded when TypeFieldEntry is instantiated,
       * and the children of the root category could not be retrieved.
       */
      if (CATEGORIES.equals(optionName) && rootCategoryId != null && super.getOption(CATEGORIES) == null) {
        Category root = Channel.getChannel().getCategory(rootCategoryId);
        if (root != null) {
          category(displayRoot ? new Category[] { root } : root.getChildrenSet().toArray(new Category[] {}));
        }
      }
    }

    return super.getOption(optionName);
  }

  private void resolveSuperTypeClass(String optionName, boolean isChannelReady) {
    String superType = super.getOption(SUPER_TYPE);
    if (SUPER_TYPE_CLASS.equals(optionName) && superType != null && super.getOption(SUPER_TYPE_CLASS) == null) {
      try {
        Class<?> superTypeClass = isChannelReady ? Channel.getChannel().getClass(superType) : Class.forName(superType);
        addOption(SUPER_TYPE_CLASS, superTypeClass);

        // Check if the type has required fields. If true, disable the file input to force the use of the "Publication Chooser"
        TypeFieldEntry[] typeFieldEntries = Channel.getChannel().getAllTypeFieldEntries(superTypeClass);
        if (typeFieldEntries != null) {
          for (TypeFieldEntry typeFieldEntry : typeFieldEntries) {
            if (!"title".equals(typeFieldEntry.getName()) && typeFieldEntry.isRequired()) {
              disableFileInput();
              break;
            }
          }
        }
      } catch (ClassNotFoundException e) {
        LOG.warn("Could not resolve super type '" + superType + "': " + e.getMessage(), e);
      }
    }
  }

  @Override
  public void resolveOptions(TypeFieldEntry tfe) {
    super.resolveOptions(tfe);

    String[] categoryIds = getOption(CATEGORIES);
    if (categoryIds == null) {
      this.rootCategoryId = tfe.getRoot();
      this.displayRoot = tfe.isDisplayRoot();
    }

    if (tfe.isLink()) {
      superType(tfe.getTypeShort());
    }
  }

  @Override
  protected String getChooserName() {
    return "publication";
  }

  @Override
  protected String getChooserOption(Map<String, String[]> customOptionMap) {
    StringBuilder chooserOption = new StringBuilder();

    String superType = getOption(SUPER_TYPE);
    if (superType != null) {
      ControlUtils.appendQueryStringParam(chooserOption, "super", superType);
    }

    Integer pstatus = getOption(PSTATUS);
    if (pstatus != null) {
      ControlUtils.appendQueryStringParam(chooserOption, "pstatus", pstatus.toString());
    }

    String[] categorieIds = getOption(CATEGORIES);
    if (categorieIds != null) {
      for (String categoryId : categorieIds) {
        // We add "cid" and "cids" because it's not consistent between the JSP of autocomplete and the JSP of the chooser window
        ControlUtils.appendQueryStringParam(chooserOption, "cid", categoryId); // 'cid' parameter for chooser
        ControlUtils.appendQueryStringParam(chooserOption, "cids", categoryId); // 'cids' parameter for autocomplete
      }
    }

    Boolean noWorkspaceFilter = getOption(NO_WORKSPACE_FILTER);
    if (noWorkspaceFilter != null) {
      ControlUtils.appendQueryStringParam(chooserOption, "workspaceFilter", String.valueOf(!noWorkspaceFilter));
    }

    String[] documentKinds = getOption(DOCUMENT_KINDS);
    if (documentKinds != null) {
      List<String> docKindsList = getDocumentKindsList();

      for (String documentKind : documentKinds) {
        if (docKindsList.contains(documentKind)) {
          ControlUtils.appendQueryStringParam(chooserOption, "documentKinds", documentKind);
        } else {
          String kindsList = StringUtils.join(docKindsList.toArray(), ", ");
          LOG.warn("Document Kind '" + documentKind + "' is not valid. You must use one of the following values: " + kindsList);
        }
      }
    }

    Boolean exactType = getOption(EXACT_TYPE);
    if (exactType != null) {
      ControlUtils.appendQueryStringParam(chooserOption, "exactType", exactType.toString());
    }

    return chooserOption.toString();
  }

  @Override
  protected String getChooserIcon() {
    Class<?> superTypeClass = getOption(SUPER_TYPE_CLASS);
    return superTypeClass != null ? JcmsUtil.getClassIcon(superTypeClass) : super.getChooserIcon();
  }

  @Override
  protected boolean useCurrentWorkspace() {
    return true;
  }

  @Override
  public PublicationSettings clone() throws CloneNotSupportedException {
    PublicationSettings cloned = super.clone();
    cloned.rootCategoryId = rootCategoryId;
    cloned.displayRoot = displayRoot;
    return cloned;
  }

  public PublicationSettings superType(Class<?> clazz) {
    return superType(clazz.getCanonicalName());
  }

  public PublicationSettings superType(String className) {
    addOption(SUPER_TYPE, className);
    return getThis();
  }

  public PublicationSettings pstatus(int pstatus) {
    addOption(PSTATUS, pstatus);
    return getThis();
  }

  public PublicationSettings category(String[] categoryIds) {
    if (categoryIds != null && categoryIds.length > 0) {
      addOption(CATEGORIES, categoryIds);
    }
    return getThis();
  }

  public PublicationSettings category(Category[] categories) {
    return category(toStringArray(categories));
  }

  public PublicationSettings category(Collection<Category> categories) {
    return category(categories != null && !categories.isEmpty() ? categories.toArray(new Category[] {}) : null);
  }

  public PublicationSettings category(Category category, Category... otherCategories) {
    return category(toStringArray(category, otherCategories));
  }

  public PublicationSettings category(String categoryId, String... otherCategoryIds) {
    return category(toStringArray(categoryId, otherCategoryIds));
  }

  public PublicationSettings noWorkspaceFilter() {
    addOption(NO_WORKSPACE_FILTER, true);
    return getThis();
  }

  public PublicationSettings documentKinds(String[] documentKinds) {
    if (documentKinds != null && documentKinds.length > 0) {
      addOption(DOCUMENT_KINDS, documentKinds);
    }
    return getThis();
  }

  public PublicationSettings documentKinds(Collection<String> documentKinds) {
    return documentKinds(documentKinds != null && !documentKinds.isEmpty() ? documentKinds.toArray(new String[] {}) : null);
  }

  public PublicationSettings documentKinds(String documentKinds, String... otherDocumentKinds) {
    return documentKinds(toStringArray(documentKinds, otherDocumentKinds));
  }

  public static List<String> getDocumentKindsList() {
    if (documentKindsList == null) {
      documentKindsList = new ArrayList<String>();

      Channel channel = Channel.getChannel();
      LangProperties groupProperties = channel != null ? channel.getProperties("file-document.kinds-group.") : null;
      if (groupProperties != null) {
        for (String property : groupProperties.keySet()) {
          List<String> kindList = Util.splitToList(channel.getProperty(property), ", ");
          if (kindList != null && !kindList.isEmpty()) {
            documentKindsList.addAll(kindList);
          }
        }
      }
    }
    return documentKindsList;
  }

  public PublicationSettings exactType() {
    addOption(EXACT_TYPE, true);
    return getThis();
  }

  public PublicationSettings disableFileInput() {
    addOption(FILE_INPUT_DISABLED, true);
    return getThis();
  }
  
  public PublicationSettings enableFileInput() {
    addOption(FILE_INPUT_ENABLED, true);
    return getThis();
  }

  public PublicationSettings disableDropdown() {
    addOption(DROPDOWN_DISABLED, true);
    return getThis();
  }

}
